from __future__ import annotations

import json
from collections import OrderedDict
from collections.abc import Iterable
from datetime import timedelta
from typing import Any, Optional, Union

import yaml
from yaml.parser import ParserError
from yaml.scanner import ScannerError

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel, CloudFormationModel
from moto.core.utils import iso_8601_datetime_with_milliseconds, utcnow
from moto.moto_api._internal import mock_random
from moto.organizations.models import OrganizationsBackend, organizations_backends
from moto.sns.models import sns_backends
from moto.utilities.utils import get_partition

from .custom_model import CustomModel
from .exceptions import (
    AlreadyExistsException,
    StackSetNotEmpty,
    StackSetNotFoundException,
    ValidationError,
)
from .parsing import Export, OutputMap, ResourceMap
from .utils import (
    generate_changeset_id,
    generate_stack_id,
    generate_stackset_arn,
    generate_stackset_id,
    get_stack_from_s3_url,
    validate_create_change_set,
    validate_template_cfn_lint,
    yaml_tag_constructor,
)


class StackSet(BaseModel):
    def __init__(
        self,
        stackset_id: str,
        account_id: str,
        name: str,
        template: str,
        region: str,
        description: Optional[str],
        parameters: dict[str, str],
        permission_model: str,
        tags: Optional[dict[str, str]],
        admin_role: Optional[str],
        execution_role: Optional[str],
    ):
        self.id = stackset_id
        self.arn = generate_stackset_arn(stackset_id, region, account_id)
        self.name = name
        self.template = template
        self.description = description
        self.parameters = parameters
        self.tags = tags
        self.administration_role_arn = (
            admin_role
            or f"arn:{get_partition(region)}:iam::{account_id}:role/AWSCloudFormationStackSetAdministrationRole"
        )
        self.execution_role_name = (
            execution_role or "AWSCloudFormationStackSetExecutionRole"
        )
        self.status = "ACTIVE"
        self.instances = StackInstances(
            account_id=account_id,
            region=region,
            template=template,
            parameters=parameters,
            stackset_id=self.id,
            stackset_name=self.name,
        )
        self.stack_instances = self.instances.stack_instances
        self.operations: list[dict[str, Any]] = []
        self.permission_model = permission_model or "SELF_MANAGED"

    @property
    def template_body(self) -> str:
        return self.template

    @property
    def managed_execution(self) -> dict[str, Any]:
        return {"Active": False}

    def _create_operation(
        self,
        operation_id: str,
        action: str,
        status: str,
        accounts: Optional[list[str]] = None,
        regions: Optional[list[str]] = None,
    ) -> dict[str, Any]:
        accounts = accounts or []
        regions = regions or []
        operation = {
            "OperationId": operation_id,
            "Action": action,
            "Status": status,
            "CreationTimestamp": utcnow(),
            "EndTimestamp": utcnow() + timedelta(minutes=2),
            "Instances": [
                {account: region} for account in accounts for region in regions
            ],
        }

        self.operations += [operation]
        return operation

    def get_operation(self, operation_id: str) -> dict[str, Any]:
        for operation in self.operations:
            if operation_id == operation["OperationId"]:
                return operation
        raise ValidationError(operation_id)

    def update_operation(self, operation_id: str, status: str) -> str:
        operation = self.get_operation(operation_id)
        operation["Status"] = status
        return operation_id

    def delete(self) -> None:
        self.status = "DELETED"

    def update(
        self,
        template: str,
        description: str,
        parameters: dict[str, str],
        tags: dict[str, str],
        admin_role: str,
        execution_role: str,
        accounts: list[str],
        regions: list[str],
        operation_id: str,
    ) -> dict[str, Any]:
        self.template = template or self.template
        self.description = description if description is not None else self.description
        self.parameters = parameters or self.parameters
        self.tags = tags or self.tags
        self.administration_role_arn = admin_role or self.administration_role_arn
        self.execution_role_name = execution_role or self.execution_role_name

        if accounts and regions:
            self.update_instances(accounts, regions, self.parameters)  # type: ignore[arg-type]

        operation = self._create_operation(
            operation_id=operation_id,
            action="UPDATE",
            status="SUCCEEDED",
            accounts=accounts,
            regions=regions,
        )
        return operation

    def create_stack_instances(
        self,
        accounts: list[str],
        regions: list[str],
        deployment_targets: Optional[dict[str, Any]],
        parameters: list[dict[str, Any]],
    ) -> str:
        if self.permission_model == "SERVICE_MANAGED":
            if not deployment_targets:
                raise ValidationError(
                    message="StackSets with SERVICE_MANAGED permission model can only have OrganizationalUnit as target"
                )
            elif "OrganizationalUnitIds" not in deployment_targets:
                raise ValidationError(message="OrganizationalUnitIds are required")
        if self.permission_model == "SELF_MANAGED":
            if deployment_targets and "OrganizationalUnitIds" in deployment_targets:
                raise ValidationError(
                    message="StackSets with SELF_MANAGED permission model can only have accounts as target"
                )
        operation_id = str(mock_random.uuid4())
        if not parameters:
            parameters = self.parameters  # type: ignore[assignment]

        self.instances.create_instances(
            accounts,
            regions,
            parameters,
            deployment_targets or {},
            permission_model=self.permission_model,
        )
        self._create_operation(
            operation_id=operation_id,
            action="CREATE",
            status="SUCCEEDED",
            accounts=accounts,
            regions=regions,
        )
        return operation_id

    def delete_stack_instances(self, accounts: list[str], regions: list[str]) -> None:
        operation_id = str(mock_random.uuid4())

        self.instances.delete(accounts, regions)

        self._create_operation(
            operation_id=operation_id,
            action="DELETE",
            status="SUCCEEDED",
            accounts=accounts,
            regions=regions,
        )

    def update_instances(
        self, accounts: list[str], regions: list[str], parameters: list[dict[str, Any]]
    ) -> dict[str, Any]:
        operation_id = str(mock_random.uuid4())

        self.instances.update(accounts, regions, parameters)
        operation = self._create_operation(
            operation_id=operation_id,
            action="UPDATE",
            status="SUCCEEDED",
            accounts=accounts,
            regions=regions,
        )
        return operation


class StackInstance(BaseModel):
    def __init__(
        self,
        account_id: str,
        region_name: str,
        stackset_id: str,
        stack_name: str,
        name: str,
        template: str,
        parameters: Optional[list[dict[str, Any]]],
        permission_model: str,
    ):
        self.account_id = account_id
        self.region_name = region_name
        self.stackset_id = stackset_id
        self.stack_name = stack_name
        self.name = name
        self.template = template
        self.parameters = parameters or []
        self.permission_model = permission_model

        # Incoming parameters can be in two formats: {key: value} or [{"": key, "": value}, ..]
        if isinstance(parameters, dict):
            params = parameters
        elif isinstance(parameters, list):
            params = {p["ParameterKey"]: p["ParameterValue"] for p in parameters}

        if permission_model == "SELF_MANAGED":
            self.stack = cloudformation_backends[account_id][region_name].create_stack(
                name=f"StackSet:{name}", template=template, parameters=params
            )
        else:
            stack_id = generate_stack_id(
                "hiddenstackfor" + self.name, self.region_name, self.account_id
            )
            self.stack = Stack(
                stack_id=stack_id,
                name=self.name,
                template=self.template,
                parameters=params,
                account_id=self.account_id,
                region_name=self.region_name,
                notification_arns=[],
                tags=None,
                role_arn=None,
                cross_stack_resources={},
                enable_termination_protection=False,
            )
            self.stack.create_resources()

    def delete(self) -> None:
        if self.permission_model == "SELF_MANAGED":
            cloudformation_backends[self.account_id][self.region_name].delete_stack(
                self.stack.name
            )
        else:
            # Our stack is hidden - we have to delete it manually
            self.stack.delete()

    def to_dict(self) -> dict[str, Any]:
        return {
            "StackId": generate_stack_id(
                self.stack_name, self.region_name, self.account_id
            ),
            "StackSetId": self.stackset_id,
            "Region": self.region_name,
            "Account": self.account_id,
            "Status": "CURRENT",
            "ParameterOverrides": self.parameters,
            "StackInstanceStatus": {"DetailedStatus": "SUCCEEDED"},
        }


class StackInstances(BaseModel):
    def __init__(
        self,
        account_id: str,
        region: str,
        template: str,
        parameters: dict[str, str],
        stackset_id: str,
        stackset_name: str,
    ):
        self.account_id = account_id
        self.partition = get_partition(region)
        self.template = template
        self.parameters = parameters or {}
        self.stackset_id = stackset_id
        self.stack_name = f"StackSet-{stackset_id}"
        self.stackset_name = stackset_name
        self.stack_instances: list[StackInstance] = []

    @property
    def org_backend(self) -> OrganizationsBackend:
        return organizations_backends[self.account_id][self.partition]

    def create_instances(
        self,
        accounts: list[str],
        regions: list[str],
        parameters: Optional[list[dict[str, Any]]],
        deployment_targets: dict[str, Any],
        permission_model: str,
    ) -> list[dict[str, Any]]:
        targets: list[tuple[str, str]] = []
        all_accounts = self.org_backend.accounts
        requested_ous = deployment_targets.get("OrganizationalUnitIds", [])
        child_ous = [
            ou.id for ou in self.org_backend.ou if ou.parent_id in requested_ous
        ]
        for region in regions:
            for account in accounts:
                targets.append((region, account))
            for ou_id in requested_ous + child_ous:
                for acnt in all_accounts:
                    if acnt.parent_id == ou_id:
                        targets.append((region, acnt.id))

        new_instances = []
        for region, account in targets:
            instance = StackInstance(
                account_id=account,
                region_name=region,
                stackset_id=self.stackset_id,
                stack_name=self.stack_name,
                name=self.stackset_name,
                template=self.template,
                parameters=parameters,
                permission_model=permission_model,
            )
            new_instances.append(instance)
        self.stack_instances += new_instances
        return [i.to_dict() for i in new_instances]

    def update(
        self,
        accounts: list[str],
        regions: list[str],
        parameters: Optional[list[dict[str, Any]]],
    ) -> Any:
        for account in accounts:
            for region in regions:
                instance = self.get_instance(account, region)
                instance.parameters = parameters or []

    def delete(self, accounts: list[str], regions: list[str]) -> None:
        to_delete = [
            i
            for i in self.stack_instances
            if i.region_name in regions and i.account_id in accounts
        ]
        for instance in to_delete:
            instance.delete()
            self.stack_instances.remove(instance)

    def get_instance(self, account: str, region: str) -> StackInstance:  # type: ignore[return]
        for i, instance in enumerate(self.stack_instances):
            if instance.region_name == region and instance.account_id == account:
                return self.stack_instances[i]


class Stack(CloudFormationModel):
    class Meta:
        serialization_aliases = {
            "Parameters": "parameter_list",
        }

    def __init__(
        self,
        stack_id: str,
        name: str,
        template: Union[str, dict[str, Any]],
        parameters: dict[str, str],
        account_id: str,
        region_name: str,
        notification_arns: Optional[list[str]] = None,
        tags: Optional[dict[str, str]] = None,
        role_arn: Optional[str] = None,
        cross_stack_resources: Optional[dict[str, Export]] = None,
        enable_termination_protection: Optional[bool] = False,
        timeout_in_mins: Optional[int] = None,
        stack_policy_body: Optional[str] = None,
    ):
        self.stack_id = stack_id
        self.name = name
        self.account_id = account_id
        self.template = template
        self.template_dict: dict[str, Any]
        if template != {}:
            self._parse_template()
            self.description = self.template_dict.get("Description")
        else:
            self.template_dict = {}
            self.description = None
        self.parameters = parameters
        self.region_name = region_name
        self.notification_arns = notification_arns if notification_arns else []
        self.role_arn = role_arn
        self.tags = tags if tags else {}
        self.events: list[Event] = []
        self.timeout_in_minutes = timeout_in_mins
        self.policy = stack_policy_body or ""

        self.cross_stack_resources: dict[str, Export] = cross_stack_resources or {}
        self.enable_termination_protection: bool = (
            enable_termination_protection or False
        )
        self.resource_map = self._create_resource_map()

        self.custom_resources: dict[str, CustomModel] = {}

        self.output_map = self._create_output_map()
        self.creation_time = utcnow()
        self.status = "CREATE_PENDING"
        self.disable_rollback = False

    def has_template(self, other_template: str) -> bool:
        self._parse_template()
        return self.template_dict == self.parse_template(other_template)

    def has_parameters(self, other_parameters: dict[str, Any]) -> bool:
        return self.parameters == other_parameters

    def _create_resource_map(self) -> ResourceMap:
        resource_map = ResourceMap(
            self.stack_id,
            self.name,
            self.parameters,
            self.tags,
            account_id=self.account_id,
            region_name=self.region_name,
            template=self.template_dict,
            cross_stack_resources=self.cross_stack_resources,
        )
        resource_map.load()
        return resource_map

    def _create_output_map(self) -> OutputMap:
        return OutputMap(self.resource_map, self.template_dict, self.stack_id)

    def _add_stack_event(
        self,
        resource_status: str,
        resource_status_reason: Optional[str] = None,
        resource_properties: Optional[str] = None,
    ) -> None:
        event = Event(
            stack_id=self.stack_id,
            stack_name=self.name,
            logical_resource_id=self.name,
            physical_resource_id=self.stack_id,
            resource_type="AWS::CloudFormation::Stack",
            resource_status=resource_status,
            resource_status_reason=resource_status_reason,
            resource_properties=resource_properties,
        )

        event.sendToSns(self.account_id, self.region_name, self.notification_arns)
        self.events.append(event)

    def _parse_template(self) -> None:
        self.template_dict = self.parse_template(self.template)  # type: ignore[arg-type]

    @staticmethod
    def parse_template(template: str) -> dict[str, Any]:  # type: ignore[misc]
        yaml.add_multi_constructor("", yaml_tag_constructor)
        try:
            return yaml.load(template, Loader=yaml.Loader)
        except (ParserError, ScannerError):
            return json.loads(template)

    @property
    def parameter_list(self) -> list[dict[str, Any]]:  # type: ignore[misc]
        parameters = [
            {
                "ParameterKey": k,
                "ParameterValue": v
                if v not in self.resource_map.no_echo_parameter_keys
                else "****",
            }
            for k, v in self.resource_map.resolved_parameters.items()
        ]
        return parameters

    @property
    def stack_parameters(self) -> dict[str, Any]:  # type: ignore[misc]
        return self.resource_map.resolved_parameters

    @property
    def stack_resources(self) -> Iterable[type[CloudFormationModel]]:
        return self.resource_map.values()

    @property
    def outputs(self) -> Optional[list[dict[str, Any]]]:
        def get_export_name(output_value: Any) -> Optional[str]:
            for export in self.exports:
                if output_value == export.value:
                    return export.name
            return None

        outputs = [
            {
                "OutputKey": o.key,
                "OutputValue": o.value,
                "Description": o.description,
                "ExportName": get_export_name(o.value),
            }
            for o in self.output_map.values()
            if o
        ]
        return outputs if outputs else None

    @property
    def exports(self) -> list[Export]:
        return self.output_map.exports

    @property
    def template_description(self) -> Optional[str]:
        return self.template_dict.get("Description")

    def add_custom_resource(self, custom_resource: CustomModel) -> None:
        self.custom_resources[custom_resource.logical_id] = custom_resource

    def get_custom_resource(self, custom_resource: str) -> CustomModel:
        return self.custom_resources[custom_resource]

    def create_resources(self) -> None:
        self.status = "CREATE_IN_PROGRESS"
        all_resources_ready = self.resource_map.create(self.template_dict)
        # Set the description of the stack
        self.description = self.template_dict.get("Description")
        if all_resources_ready:
            self.mark_creation_complete()

    def verify_readiness(self) -> None:
        if self.resource_map.creation_complete():
            self.mark_creation_complete()

    def mark_creation_complete(self) -> None:
        self.status = "CREATE_COMPLETE"
        self._add_stack_event("CREATE_COMPLETE")

    def update(
        self,
        template: str,
        role_arn: Optional[str] = None,
        parameters: Optional[dict[str, Any]] = None,
        tags: Optional[dict[str, str]] = None,
    ) -> None:
        self._add_stack_event(
            "UPDATE_IN_PROGRESS", resource_status_reason="User Initiated"
        )
        self.template = template
        self._parse_template()
        self.resource_map.update(self.template_dict, parameters)
        self.output_map = self._create_output_map()
        self._add_stack_event("UPDATE_COMPLETE")
        self.status = "UPDATE_COMPLETE"
        self.role_arn = role_arn
        if parameters:
            self.parameters = parameters
        # only overwrite tags if passed
        if tags is not None:
            self.tags = tags
            # TODO: update tags in the resource map

    def delete(self) -> None:
        self._add_stack_event(
            "DELETE_IN_PROGRESS", resource_status_reason="User Initiated"
        )
        self.resource_map.delete()
        self._add_stack_event("DELETE_COMPLETE")
        self.status = "DELETE_COMPLETE"

    @staticmethod
    def cloudformation_type() -> str:
        return "AWS::CloudFormation::Stack"

    @classmethod
    def has_cfn_attr(cls, attr: str) -> bool:
        return True

    @property
    def physical_resource_id(self) -> str:
        return self.name

    @classmethod
    def create_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: dict[str, Any],
        account_id: str,
        region_name: str,
        **kwargs: Any,
    ) -> Stack:
        cf_backend: CloudFormationBackend = cloudformation_backends[account_id][
            region_name
        ]
        properties = cloudformation_json["Properties"]

        template_body = get_stack_from_s3_url(
            properties["TemplateURL"],
            account_id=account_id,
            partition=get_partition(region_name),
        )
        parameters = properties.get("Parameters", {})

        return cf_backend.create_stack(
            name=resource_name, template=template_body, parameters=parameters
        )

    @classmethod
    def update_from_cloudformation_json(  # type: ignore[misc]
        cls,
        original_resource: Any,
        new_resource_name: str,
        cloudformation_json: Any,
        account_id: str,
        region_name: str,
    ) -> Stack:
        cls.delete_from_cloudformation_json(
            original_resource.name, cloudformation_json, account_id, region_name
        )
        return cls.create_from_cloudformation_json(
            new_resource_name, cloudformation_json, account_id, region_name
        )

    @classmethod
    def delete_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: dict[str, Any],
        account_id: str,
        region_name: str,
    ) -> None:
        cf_backend: CloudFormationBackend = cloudformation_backends[account_id][
            region_name
        ]
        cf_backend.delete_stack(resource_name)


class Change(BaseModel):
    def __init__(self, action: str, logical_resource_id: str, resource_type: str):
        self.action = action
        self.logical_resource_id = logical_resource_id
        self.resource_type = resource_type


class ChangeSet(BaseModel):
    def __init__(
        self,
        change_set_type: str,
        change_set_id: str,
        change_set_name: str,
        stack: Stack,
        template: str,
        parameters: dict[str, str],
        description: str,
        notification_arns: Optional[list[str]] = None,
        tags: Optional[dict[str, str]] = None,
        role_arn: Optional[str] = None,
    ):
        self.change_set_type = change_set_type
        self.change_set_id = change_set_id
        self.change_set_name = change_set_name

        self.stack = stack
        self.stack_id = self.stack.stack_id
        self.stack_name = self.stack.name
        self.notification_arns = notification_arns
        self.description = description
        self.tags = tags
        self.role_arn = role_arn
        self.template = template
        self.parameters = parameters
        self._parse_template()

        self.creation_time = utcnow()
        self.changes = self.diff()

        self.status: Optional[str] = None
        self.execution_status: Optional[str] = None
        self.status_reason: Optional[str] = None

    def _parse_template(self) -> None:
        yaml.add_multi_constructor("", yaml_tag_constructor)
        try:
            self.template_dict = yaml.load(self.template, Loader=yaml.Loader)
        except (ParserError, ScannerError):
            self.template_dict = json.loads(self.template)

    def diff(self) -> list[Change]:
        changes = []
        resources_by_action = self.stack.resource_map.build_change_set_actions(
            self.template_dict
        )
        for action, resources in resources_by_action.items():
            for resource_name, resource in resources.items():
                changes.append(
                    Change(
                        action=action,
                        logical_resource_id=resource_name,
                        resource_type=resource["ResourceType"],
                    )
                )
        return changes

    def apply(self) -> None:
        self.stack.resource_map.update(self.template_dict, self.parameters)


class Event(BaseModel):
    def __init__(
        self,
        stack_id: str,
        stack_name: str,
        logical_resource_id: str,
        physical_resource_id: str,
        resource_type: str,
        resource_status: str,
        resource_status_reason: Optional[str],
        resource_properties: Optional[str],
    ):
        self.stack_id = stack_id
        self.stack_name = stack_name
        self.logical_resource_id = logical_resource_id
        self.physical_resource_id = physical_resource_id
        self.resource_type = resource_type
        self.resource_status = resource_status
        self.resource_status_reason = resource_status_reason
        self.resource_properties = resource_properties
        self.timestamp = utcnow()
        self.event_id = mock_random.uuid4()
        self.client_request_token = None

    def sendToSns(
        self, account_id: str, region: str, sns_topic_arns: list[str]
    ) -> None:
        message = f"""StackId='{self.stack_id}'
Timestamp='{iso_8601_datetime_with_milliseconds(self.timestamp)}'
EventId='{self.event_id}'
LogicalResourceId='{self.logical_resource_id}'
Namespace='{account_id}'
ResourceProperties='{self.resource_properties}'
ResourceStatus='{self.resource_status}'
ResourceStatusReason='{self.resource_status_reason}'
ResourceType='{self.resource_type}'
StackName='{self.stack_name}'
ClientRequestToken='{self.client_request_token}'"""

        for sns_topic_arn in sns_topic_arns:
            sns_backends[account_id][region].publish(
                message, subject="AWS CloudFormation Notification", arn=sns_topic_arn
            )


def filter_stacks(
    all_stacks: list[Stack], status_filter: Optional[list[str]]
) -> list[Stack]:
    filtered_stacks = []
    if not status_filter:
        return all_stacks
    for stack in all_stacks:
        if stack.status in status_filter:
            filtered_stacks.append(stack)
    return filtered_stacks


class CloudFormationBackend(BaseBackend):
    """
    CustomResources are supported when running Moto in ServerMode.
    Because creating these resources involves running a Lambda-function that informs the MotoServer about the status of the resources, the MotoServer has to be reachable for outside connections.
    This means it has to run inside a Docker-container, or be started using `moto_server -h 0.0.0.0`.
    """

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.stacks: dict[str, Stack] = OrderedDict()
        self.stacksets: dict[str, StackSet] = OrderedDict()
        self.deleted_stacks: dict[str, Stack] = {}
        self.exports: dict[str, Export] = OrderedDict()
        self.change_sets: dict[str, ChangeSet] = OrderedDict()

    @staticmethod
    def default_vpc_endpoint_service(
        service_region: str, zones: list[str]
    ) -> list[dict[str, str]]:
        """Default VPC endpoint service."""
        return BaseBackend.default_vpc_endpoint_service_factory(
            service_region, zones, "cloudformation", policy_supported=False
        )

    def _resolve_update_parameters(
        self,
        instance: Union[Stack, StackSet],
        incoming_params: list[dict[str, str]],
    ) -> dict[str, str]:
        parameters = {
            parameter["ParameterKey"]: parameter["ParameterValue"]
            for parameter in incoming_params
            if "ParameterValue" in parameter
        }
        previous = {
            parameter["ParameterKey"]: instance.parameters[parameter["ParameterKey"]]
            for parameter in incoming_params
            if parameter.get("UsePreviousValue", False)
        }
        parameters.update(previous)

        return parameters

    def create_stack_set(
        self,
        name: str,
        template: str,
        parameters: dict[str, str],
        tags: dict[str, str],
        permission_model: str,
        admin_role: Optional[str],
        exec_role: Optional[str],
        description: Optional[str],
    ) -> StackSet:
        """
        The following parameters are not yet implemented: StackId, AdministrationRoleARN, AutoDeployment, ExecutionRoleName, CallAs, ClientRequestToken, ManagedExecution
        """
        stackset_id = generate_stackset_id(name)
        new_stackset = StackSet(
            stackset_id=stackset_id,
            account_id=self.account_id,
            name=name,
            region=self.region_name,
            template=template,
            parameters=parameters,
            description=description,
            tags=tags,
            permission_model=permission_model,
            admin_role=admin_role,
            execution_role=exec_role,
        )
        self.stacksets[stackset_id] = new_stackset
        return new_stackset

    def describe_stack_set(self, name: str) -> StackSet:
        stacksets = self.stacksets.keys()
        if name in stacksets and self.stacksets[name].status != "DELETED":
            return self.stacksets[name]
        for stackset in stacksets:
            if (
                self.stacksets[stackset].name == name
                and self.stacksets[stackset].status != "DELETED"
            ):
                return self.stacksets[stackset]
        raise StackSetNotFoundException(name)

    def delete_stack_set(self, name: str) -> None:
        stackset_to_delete: Optional[StackSet] = None
        if name in self.stacksets:
            stackset_to_delete = self.stacksets[name]
        for stackset in self.stacksets.values():
            if stackset.name == name:
                stackset_to_delete = stackset

        if stackset_to_delete is not None:
            if stackset_to_delete.stack_instances:
                raise StackSetNotEmpty()
            # We don't remove StackSets from the list - they still show up when calling list_stack_sets
            stackset_to_delete.delete()

    def list_stack_sets(self) -> Iterable[StackSet]:
        return self.stacksets.values()

    def list_stack_set_operations(self, stackset_name: str) -> list[dict[str, Any]]:
        stackset = self.describe_stack_set(stackset_name)
        return stackset.operations

    def stop_stack_set_operation(self, stackset_name: str, operation_id: str) -> None:
        stackset = self.describe_stack_set(stackset_name)
        stackset.update_operation(operation_id, "STOPPED")

    def describe_stack_set_operation(
        self, stackset_name: str, operation_id: str
    ) -> tuple[StackSet, dict[str, Any]]:
        stackset = self.describe_stack_set(stackset_name)
        operation = stackset.get_operation(operation_id)
        return stackset, operation

    def list_stack_set_operation_results(
        self, stackset_name: str, operation_id: str
    ) -> dict[str, Any]:
        stackset = self.describe_stack_set(stackset_name)
        return stackset.get_operation(operation_id)

    def create_stack_instances(
        self,
        stackset_name: str,
        accounts: list[str],
        regions: list[str],
        parameters: list[dict[str, str]],
        deployment_targets: Optional[dict[str, Any]],
    ) -> str:
        """
        The following parameters are not yet implemented: DeploymentTargets.AccountFilterType, DeploymentTargets.AccountsUrl, OperationPreferences, CallAs
        """
        stackset = self.describe_stack_set(stackset_name)

        operation_id = stackset.create_stack_instances(
            accounts=accounts,
            regions=regions,
            deployment_targets=deployment_targets,
            parameters=parameters,
        )
        return operation_id

    def update_stack_instances(
        self,
        stackset_name: str,
        accounts: list[str],
        regions: list[str],
        parameters: list[dict[str, Any]],
    ) -> dict[str, Any]:
        """
        Calling this will update the parameters, but the actual resources are not updated
        """
        stack_set = self.describe_stack_set(stackset_name)
        return stack_set.update_instances(accounts, regions, parameters)

    def update_stack_set(
        self,
        stackset_name: str,
        template: str,
        description: str,
        parameters: list[dict[str, str]],
        tags: dict[str, str],
        admin_role: str,
        execution_role: str,
        accounts: list[str],
        regions: list[str],
        operation_id: str,
    ) -> dict[str, Any]:
        stackset = self.describe_stack_set(stackset_name)
        resolved_parameters = self._resolve_update_parameters(
            instance=stackset, incoming_params=parameters
        )
        update = stackset.update(
            template=template,
            description=description,
            parameters=resolved_parameters,
            tags=tags,
            admin_role=admin_role,
            execution_role=execution_role,
            accounts=accounts,
            regions=regions,
            operation_id=operation_id,
        )
        return update

    def delete_stack_instances(
        self, stackset_name: str, accounts: list[str], regions: list[str]
    ) -> StackSet:
        """
        The following parameters are not  yet implemented: DeploymentTargets, OperationPreferences, RetainStacks, OperationId, CallAs
        """
        stackset = self.describe_stack_set(stackset_name)
        stackset.delete_stack_instances(accounts, regions)
        return stackset

    def create_stack(
        self,
        name: str,
        template: str,
        parameters: dict[str, Any],
        notification_arns: Optional[list[str]] = None,
        tags: Optional[dict[str, str]] = None,
        role_arn: Optional[str] = None,
        enable_termination_protection: Optional[bool] = False,
        timeout_in_mins: Optional[int] = None,
        stack_policy_body: Optional[str] = None,
    ) -> Stack:
        """
        The functionality behind EnableTerminationProtection is not yet implemented.
        """
        if name in [stack.name for stack in self.stacks.values()]:
            raise AlreadyExistsException(f"Stack {name} already exists")
        stack_id = generate_stack_id(name, self.region_name, self.account_id)
        new_stack = Stack(
            stack_id=stack_id,
            name=name,
            template=template,
            parameters=parameters,
            account_id=self.account_id,
            region_name=self.region_name,
            notification_arns=notification_arns,
            tags=tags,
            role_arn=role_arn,
            cross_stack_resources=self.exports,
            enable_termination_protection=enable_termination_protection,
            timeout_in_mins=timeout_in_mins,
            stack_policy_body=stack_policy_body,
        )
        self.stacks[stack_id] = new_stack
        self._validate_export_uniqueness(new_stack)
        for export in new_stack.exports:
            self.exports[export.name] = export
        new_stack._add_stack_event(
            "CREATE_IN_PROGRESS", resource_status_reason="User Initiated"
        )
        new_stack.create_resources()
        return new_stack

    def create_change_set(
        self,
        stack_name: str,
        change_set_name: str,
        template: str,
        parameters: dict[str, str],
        description: str,
        change_set_type: str,
        notification_arns: Optional[list[str]] = None,
        tags: Optional[dict[str, str]] = None,
        role_arn: Optional[str] = None,
    ) -> tuple[str, str]:
        validate_create_change_set(change_set_name)
        if change_set_type == "UPDATE":
            for stack in self.stacks.values():
                if stack.name == stack_name:
                    break
            else:
                raise ValidationError(stack_name)
        else:
            stack_id = generate_stack_id(stack_name, self.region_name, self.account_id)
            stack = Stack(
                stack_id=stack_id,
                name=stack_name,
                template={},
                parameters=parameters,
                account_id=self.account_id,
                region_name=self.region_name,
                notification_arns=notification_arns,
                tags=tags,
                role_arn=role_arn,
            )
            self.stacks[stack_id] = stack
            stack.status = "REVIEW_IN_PROGRESS"
            stack._add_stack_event(
                "REVIEW_IN_PROGRESS", resource_status_reason="User Initiated"
            )

        change_set_id = generate_changeset_id(
            change_set_name, self.region_name, self.account_id
        )

        new_change_set = ChangeSet(
            change_set_type=change_set_type,
            change_set_id=change_set_id,
            change_set_name=change_set_name,
            stack=stack,
            template=template,
            parameters=parameters,
            description=description,
            notification_arns=notification_arns,
            tags=tags,
            role_arn=role_arn,
        )
        if (
            change_set_type == "UPDATE"
            and stack.has_template(template)
            and stack.has_parameters(parameters)
        ):
            # Nothing has changed - mark it as such
            new_change_set.status = "FAILED"
            new_change_set.execution_status = "UNAVAILABLE"
            new_change_set.status_reason = "The submitted information didn't contain changes. Submit different information to create a change set."
        else:
            new_change_set.status = "CREATE_COMPLETE"
            new_change_set.execution_status = "AVAILABLE"
        self.change_sets[change_set_id] = new_change_set
        return change_set_id, stack.stack_id

    def delete_change_set(self, change_set_name: str) -> None:
        if change_set_name in self.change_sets:
            # This means arn was passed in
            del self.change_sets[change_set_name]
        else:
            for cs in self.change_sets:
                if self.change_sets[cs].change_set_name == change_set_name:
                    to_delete = cs
                    break
            del self.change_sets[to_delete]

    def describe_change_set(self, change_set_name: str) -> Optional[ChangeSet]:
        change_set = None
        if change_set_name in self.change_sets:
            # This means arn was passed in
            change_set = self.change_sets[change_set_name]
        else:
            for cs in self.change_sets:
                if self.change_sets[cs].change_set_name == change_set_name:
                    change_set = self.change_sets[cs]
        if change_set is None:
            raise ValidationError(change_set_name)
        return change_set

    def execute_change_set(
        self, change_set_name: str, stack_name: Optional[str] = None
    ) -> None:
        if change_set_name in self.change_sets:
            # This means arn was passed in
            change_set = self.change_sets[change_set_name]
        else:
            for cs in self.change_sets:
                if self.change_sets[cs].change_set_name == change_set_name:
                    change_set = self.change_sets[cs]

        if change_set is None:
            raise ValidationError(stack_name)

        stack = self.stacks[change_set.stack_id]
        # TODO: handle execution errors and implement rollback
        if change_set.change_set_type == "CREATE":
            stack._add_stack_event(
                "CREATE_IN_PROGRESS", resource_status_reason="User Initiated"
            )
            change_set.apply()
            stack._add_stack_event("CREATE_COMPLETE")
        else:
            stack._add_stack_event("UPDATE_IN_PROGRESS")
            change_set.apply()
            stack._add_stack_event("UPDATE_COMPLETE")

        # set the execution status of the changeset
        change_set.execution_status = "EXECUTE_COMPLETE"

        # set the status of the stack
        stack.status = f"{change_set.change_set_type}_COMPLETE"
        stack.template = change_set.template

    def describe_stacks(self, name_or_stack_id: str) -> list[Stack]:
        stacks = self.stacks.values()
        if name_or_stack_id:
            for stack in stacks:
                if stack.name == name_or_stack_id or stack.stack_id == name_or_stack_id:
                    return [stack]
            if self.deleted_stacks:
                deleted_stacks = self.deleted_stacks.values()
                for stack in deleted_stacks:
                    if stack.stack_id == name_or_stack_id:
                        return [stack]
            raise ValidationError(name_or_stack_id)
        else:
            return list(stacks)

    def describe_stack_instance(
        self, stack_set_name: str, account_id: str, region: str
    ) -> dict[str, Any]:
        stack_set = self.describe_stack_set(stack_set_name)
        return stack_set.instances.get_instance(account_id, region).to_dict()

    def list_stack_instances(self, stackset_name: str) -> list[dict[str, Any]]:
        """
        Pagination is not yet implemented.
        The parameters StackInstanceAccount/StackInstanceRegion are not yet implemented.
        """
        stack_set = self.describe_stack_set(stackset_name)
        return [i.to_dict() for i in stack_set.instances.stack_instances]

    def list_change_sets(self) -> Iterable[ChangeSet]:
        return self.change_sets.values()

    def list_stacks(self, status_filter: Optional[list[str]] = None) -> list[Stack]:
        total_stacks = list(self.stacks.values()) + list(self.deleted_stacks.values())
        return filter_stacks(total_stacks, status_filter)

    def get_stack(self, name_or_stack_id: str) -> Stack:
        all_stacks = dict(self.deleted_stacks, **self.stacks)
        if name_or_stack_id in all_stacks:
            # Lookup by stack id - deleted stacks incldued
            return all_stacks[name_or_stack_id]
        else:
            # Lookup by stack name - undeleted stacks only
            for stack in self.stacks.values():
                if stack.name == name_or_stack_id:
                    return stack
            raise ValidationError(name_or_stack_id)

    def update_stack(
        self,
        name: str,
        template: str,
        role_arn: Optional[str],
        parameters: list[dict[str, Any]],
        tags: Optional[dict[str, str]],
    ) -> Stack:
        stack = self.get_stack(name)
        resolved_parameters = self._resolve_update_parameters(
            instance=stack, incoming_params=parameters
        )
        stack.update(template, role_arn, parameters=resolved_parameters, tags=tags)
        return stack

    def get_stack_policy(self, stack_name: str) -> str:
        try:
            stack = self.get_stack(stack_name)
        except ValidationError:
            raise ValidationError(message=f"Stack: {stack_name} does not exist")
        return stack.policy

    def set_stack_policy(self, stack_name: str, policy_body: str) -> None:
        """
        Note that Moto does no validation/parsing/enforcement of this policy - we simply persist it.
        """
        try:
            stack = self.get_stack(stack_name)
        except ValidationError:
            raise ValidationError(message=f"Stack: {stack_name} does not exist")
        stack.policy = policy_body

    def describe_stack_resource(
        self, stack_name: str, logical_resource_id: str
    ) -> tuple[Stack, type[CloudFormationModel]]:
        stack = self.get_stack(stack_name)

        for stack_resource in stack.stack_resources:
            if stack_resource.logical_resource_id == logical_resource_id:  # type: ignore[attr-defined]
                return stack, stack_resource

        message = (
            f"Resource {logical_resource_id} does not exist for stack {stack_name}"
        )
        raise ValidationError(stack_name, message)

    def describe_stack_resources(
        self, stack_name: str, logical_resource_id: str
    ) -> tuple[Stack, Iterable[type[CloudFormationModel]]]:
        stack = self.get_stack(stack_name)

        if logical_resource_id is not None:
            for res in stack.stack_resources:
                if res.logical_resource_id == logical_resource_id:  # type: ignore[attr-defined]
                    return stack, [res]
            return stack, []

        return stack, stack.stack_resources

    def list_stack_resources(
        self, stack_name_or_id: str
    ) -> Iterable[type[CloudFormationModel]]:
        stack = self.get_stack(stack_name_or_id)
        return stack.stack_resources

    def delete_stack(self, name_or_stack_id: str) -> None:
        if name_or_stack_id in self.stacks:
            # Delete by stack id
            stack = self.stacks.pop(name_or_stack_id)
            export_names = [export.name for export in stack.exports]
            stack.delete()
            self.deleted_stacks[stack.stack_id] = stack
            for export_name in export_names:
                self.exports.pop(export_name)
            self.stacks.pop(name_or_stack_id, None)
        else:
            # Delete by stack name
            for stack in list(self.stacks.values()):
                if stack.name == name_or_stack_id:
                    self.delete_stack(stack.stack_id)

    def list_exports(
        self, tokenstr: Optional[str]
    ) -> tuple[list[Export], Optional[str]]:
        all_exports = list(self.exports.values())
        if tokenstr is None:
            exports = all_exports[0:100]
            next_token = "100" if len(all_exports) > 100 else None
        else:
            token = int(tokenstr)
            exports = all_exports[token : token + 100]
            next_token = str(token + 100) if len(all_exports) > token + 100 else None
        return exports, next_token

    def describe_stack_events(self, stack_name: str) -> list[Event]:
        return self.get_stack(stack_name).events

    def get_template(self, name_or_stack_id: str) -> Union[str, dict[str, Any]]:
        return self.get_stack(name_or_stack_id).template

    def validate_template(self, template: str) -> list[Any]:
        return validate_template_cfn_lint(template)

    def _validate_export_uniqueness(self, stack: Stack) -> None:
        new_stack_export_names = [x.name for x in stack.exports]
        export_names = self.exports.keys()
        if not set(export_names).isdisjoint(new_stack_export_names):
            raise ValidationError(
                stack.stack_id,
                message="Export names must be unique across a given region",
            )


cloudformation_backends = BackendDict(CloudFormationBackend, "cloudformation")



"""mediapackagev2Backend class with methods for supported APIs."""

from datetime import datetime

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.moto_api._internal import mock_random
from moto.utilities.paginator import paginate

from .exceptions import ChannelGroupNotEmpty, ChannelGroupNotFound, ChannelNotFound
from .utils import PAGINATION_MODEL


class Channel(BaseModel):
    def __init__(self, group: "ChannelGroup", channel_name: str):
        self.group = group
        self.channel_group_name = group.channel_group_name
        self.channel_name = channel_name
        self.arn = f"{group.arn}/channel/{channel_name}"

        ingest_domain1 = group.egress_domain.replace(".egress", "-1.ingest")
        ingest_domain2 = group.egress_domain.replace(".egress", "-2.ingest")
        self.ingest_endpoints = [
            {
                "Id": "1",
                "Url": f"https://{ingest_domain1}/in/v1/{group.channel_group_name}/1/{channel_name}/index",
            },
            {
                "Id": "2",
                "Url": f"https://{ingest_domain2}/in/v1/{group.channel_group_name}/2/{channel_name}/index",
            },
        ]

        self.e_tag = mock_random.get_random_hex(10)
        self.tags: dict[str, str] = {}

        self.input_switch_configuration = {"MQCSInputSwitching": False}
        self.output_header_configuration = {"PublishMQCS": False}

        self.input_type = "HLS"


class ChannelGroup(BaseModel):
    def __init__(self, account_id: str, region_name: str, channel_group_name: str):
        self.channel_group_name = channel_group_name
        self.arn = f"arn:aws:mediapackagev2:{region_name}:{account_id}:channelGroup/{channel_group_name}"
        self.egress_domain = f"{mock_random.get_random_hex(6)}.egress.{mock_random.get_random_hex(6)}.mediapackagev2.{region_name}.amazonaws.com"
        now = datetime.now()
        self.created_at = now
        self.modified_at = now
        self.e_tag = mock_random.get_random_hex(10)
        self.tags: dict[str, str] = {}

        self.channels: dict[str, Channel] = {}

    def create_channel(self, channel_name: str) -> Channel:
        channel = Channel(group=self, channel_name=channel_name)
        self.channels[channel_name] = channel
        return channel


class MediaPackagev2Backend(BaseBackend):
    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.channel_groups: dict[str, ChannelGroup] = {}

    def create_channel_group(self, channel_group_name: str) -> ChannelGroup:
        group = ChannelGroup(
            account_id=self.account_id,
            region_name=self.region_name,
            channel_group_name=channel_group_name,
        )
        self.channel_groups[channel_group_name] = group
        return group

    def create_channel(self, channel_group_name: str, channel_name: str) -> Channel:
        group = self.get_channel_group(channel_group_name)
        return group.create_channel(channel_name)

    def get_channel_group(self, channel_group_name: str) -> ChannelGroup:
        if channel_group_name not in self.channel_groups:
            raise ChannelGroupNotFound
        return self.channel_groups[channel_group_name]

    @paginate(PAGINATION_MODEL)
    def list_channel_groups(self) -> list[ChannelGroup]:
        return list(self.channel_groups.values())

    def get_channel(self, channel_group_name: str, channel_name: str) -> Channel:
        group = self.get_channel_group(channel_group_name)
        if channel_name not in group.channels:
            raise ChannelNotFound
        return group.channels[channel_name]

    def delete_channel_group(self, channel_group_name: str) -> None:
        if channel := self.channel_groups.get(channel_group_name):
            if channel.channels:
                raise ChannelGroupNotEmpty
        self.channel_groups.pop(channel_group_name, None)

    def delete_channel(self, channel_group_name: str, channel_name: str) -> None:
        group = self.get_channel_group(channel_group_name)
        group.channels.pop(channel_name, None)


mediapackagev2_backends = BackendDict(MediaPackagev2Backend, "mediapackagev2")



from __future__ import annotations

from moto.stepfunctions.parser.asl.component.common.comment import Comment
from moto.stepfunctions.parser.asl.component.common.flow.start_at import StartAt
from moto.stepfunctions.parser.asl.component.common.query_language import QueryLanguage
from moto.stepfunctions.parser.asl.component.program.states import States
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.distributed_iteration_component import (
    DistributedIterationComponent,
    DistributedIterationComponentEvalInput,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.processor_config import (
    ProcessorConfig,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iterator.distributed_iterator_worker import (
    DistributedIteratorWorker,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.job import (
    JobPool,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.parse.typed_props import TypedProps


class DistributedIteratorEvalInput(DistributedIterationComponentEvalInput):
    pass


class DistributedIterator(DistributedIterationComponent):
    @classmethod
    def from_props(cls, props: TypedProps) -> DistributedIterator:
        item_processor = cls(
            query_language=props.get(QueryLanguage) or QueryLanguage(),
            start_at=props.get(
                typ=StartAt,
                raise_on_missing=ValueError(
                    f"Missing StartAt declaration in props '{props}'."
                ),
            ),
            states=props.get(
                typ=States,
                raise_on_missing=ValueError(
                    f"Missing States declaration in props '{props}'."
                ),
            ),
            comment=props.get(Comment),
            processor_config=props.get(ProcessorConfig),
        )
        return item_processor

    def _create_worker(
        self,
        env: Environment,
        eval_input: DistributedIteratorEvalInput,
        job_pool: JobPool,
    ) -> DistributedIteratorWorker:
        return DistributedIteratorWorker(
            work_name=eval_input.state_name,
            job_pool=job_pool,
            env=env,
            parameters=eval_input.parameters,
            map_run_record=eval_input.map_run_record,
            item_selector=eval_input.item_selector,
        )



from moto.core.responses import ActionResult

from ._base_response import EC2BaseResponse


class AccountAttributes(EC2BaseResponse):
    def describe_account_attributes(self) -> ActionResult:
        attribute_defaults: dict[str, list[str]] = {
            "vpc-max-security-groups-per-interface": ["5"],
            "max-instances": ["20"],
            "supported-platforms": ["EC2", "VPC"],
            "default-vpc": ["none"],
            "max-elastic-ips": ["5"],
            "vpc-max-elastic-ips": ["5"],
        }
        result = {
            "AccountAttributes": [
                {
                    "AttributeName": k,
                    "AttributeValues": [{"AttributeValue": v} for v in values],
                }
                for k, values in attribute_defaults.items()
            ]
        }
        return ActionResult(result)



from moto.core.exceptions import JsonRESTError, ServiceException


class SesError(ServiceException):
    pass


class MessageRejectedError(SesError):
    code = "MessageRejected"


class ConfigurationSetDoesNotExist(SesError):
    code = "ConfigurationSetDoesNotExist"


class ConfigurationSetAlreadyExists(SesError):
    code = "ConfigurationSetAlreadyExists"


class EventDestinationAlreadyExists(SesError):
    code = "EventDestinationAlreadyExists"


class TemplateNameAlreadyExists(SesError):
    code = "AlreadyExists"


class ValidationError(SesError):
    code = "ValidationError"


class InvalidParameterValue(SesError):
    code = "InvalidParameterValue"


class InvalidRenderingParameterException(SesError):
    code = "InvalidRenderingParameterException"


class TemplateDoesNotExist(SesError):
    code = "TemplateDoesNotExist"


class AlreadyExists(SesError):
    code = "AlreadyExists"


class RuleSetDoesNotExist(SesError):
    code = "RuleSetDoesNotExist"


class RuleDoesNotExist(SesError):
    code = "RuleDoesNotExist"


class CannotDelete(SesError):
    code = "CannotDelete"


class InvalidS3ConfigurationException(SesError):
    code = "InvalidS3Configuration"


class InvalidSnsTopicException(SesError):
    code = "InvalidSnsTopic"


class InvalidLambdaFunctionException(SesError):
    code = "InvalidLambdaFunction"


class MissingRenderingAttributeException(SesError):
    code = "MissingRenderingAttributeException"

    def __init__(self, var: str):
        super().__init__(f"Attribute '{var}' is not present in the rendering data.")


class NotFoundException(JsonRESTError):
    code = 404

    def __init__(self, message: str):
        super().__init__("NotFoundException", message)



from typing import Final, Optional

from moto.stepfunctions.parser.api import HistoryEventType, TaskFailedEventDetails
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.common.string.string_expression import (
    StringJsonPath,
    StringSampler,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.json_path import NoSuchJsonPathError


class InputPath(EvalComponent):
    string_sampler: Final[Optional[StringSampler]]

    def __init__(self, string_sampler: Optional[StringSampler]):
        self.string_sampler = string_sampler

    def _eval_body(self, env: Environment) -> None:
        if self.string_sampler is None:
            env.stack.append({})
            return
        if isinstance(self.string_sampler, StringJsonPath):
            # JsonPaths are sampled from a given state, hence pass the state's input.
            env.stack.append(env.states.get_input())
        try:
            self.string_sampler.eval(env=env)
        except NoSuchJsonPathError as no_such_json_path_error:
            json_path = no_such_json_path_error.json_path
            cause = f"Invalid path '{json_path}' : No results for path: $['{json_path[2:]}']"
            raise FailureEventException(
                failure_event=FailureEvent(
                    env=env,
                    error_name=StatesErrorName(typ=StatesErrorNameType.StatesRuntime),
                    event_type=HistoryEventType.TaskFailed,
                    event_details=EventDetails(
                        taskFailedEventDetails=TaskFailedEventDetails(
                            error=StatesErrorNameType.StatesRuntime.to_name(),
                            cause=cause,
                        )
                    ),
                )
            )



"""Handles incoming servicecatalogappregistry requests, invokes methods, returns responses."""

import json
from urllib.parse import unquote

from moto.core.responses import BaseResponse
from moto.servicecatalogappregistry.exceptions import ValidationException

from .models import (
    Application,
    AppRegistryBackend,
    servicecatalogappregistry_backends,
)


class AppRegistryResponse(BaseResponse):
    """Handler for AppRegistry requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="servicecatalog-appregistry")

    @property
    def servicecatalogappregistry_backend(self) -> AppRegistryBackend:
        """Return backend instance specific for this region."""
        return servicecatalogappregistry_backends[self.current_account][self.region]

    def create_application(self) -> str:
        name = self._get_param("name")
        description = self._get_param("description")
        tags = self._get_param("tags")
        client_token = self._get_param("clientToken")
        application = self.servicecatalogappregistry_backend.create_application(
            name=name,
            description=description,
            tags=tags,
            client_token=client_token,
        )
        return json.dumps({"application": application.to_json()})

    def list_applications(self) -> str:
        applications = self.servicecatalogappregistry_backend.list_applications()
        json_list = []
        for app in applications:
            json_list.append(app.to_json())
        return json.dumps({"applications": json_list})

    def associate_resource(self) -> str:
        application = unquote(self._get_param("application"))
        resource_type = self._get_param("resourceType")
        resource = unquote(self._get_param("resource"))
        options = self._get_param("options")
        app = None
        app = self._find_app_by_any_value(application)
        if options is None:
            options = []
        new_resource = self.servicecatalogappregistry_backend.associate_resource(
            app.arn,
            resource_type,
            resource,
            options,
        )
        return json.dumps(new_resource)

    def list_associated_resources(self) -> str:
        application = unquote(self._get_param("application"))
        app = self._find_app_by_any_value(application)
        return_list = []
        for resource in app.associated_resources.values():
            return_list.append(resource.to_json())
        return json.dumps({"resources": return_list})

    def _find_app_by_any_value(self, search: str) -> Application:
        app = None
        if search in self.servicecatalogappregistry_backend.applications:
            app = self.servicecatalogappregistry_backend.applications[search]
        else:
            for a in self.servicecatalogappregistry_backend.applications.values():
                if search == a.id:
                    app = a
                elif search == a.name:
                    app = a
        if app is None:
            raise ValidationException
        return app

    def put_configuration(self) -> None:
        configuration = self._get_param("configuration")
        self.servicecatalogappregistry_backend.put_configuration(
            configuration=configuration,
        )

    def get_configuration(self) -> str:
        return json.dumps(
            {"configuration": self.servicecatalogappregistry_backend.configuration}
        )



from moto.core.common_types import TYPE_RESPONSE
from moto.core.responses import ActionResult, BaseResponse, EmptyResult
from moto.ec2.utils import parse_user_data
from moto.utilities.aws_headers import amz_crc32

from .models import AutoScalingBackend, autoscaling_backends


class AutoScalingResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="autoscaling")
        self.automated_parameter_parsing = True

    @property
    def autoscaling_backend(self) -> AutoScalingBackend:
        return autoscaling_backends[self.current_account][self.region]

    @amz_crc32
    def call_action(self) -> TYPE_RESPONSE:
        return super().call_action()

    def create_launch_configuration(self) -> ActionResult:
        params = self._get_params()
        user_data = parse_user_data(params.get("UserData"))
        self.autoscaling_backend.create_launch_configuration(
            name=params.get("LaunchConfigurationName"),  # type: ignore[arg-type]
            image_id=params.get("ImageId"),  # type: ignore[arg-type]
            key_name=params.get("KeyName"),
            ramdisk_id=params.get("RamdiskId"),  # type: ignore[arg-type]
            kernel_id=params.get("KernelId"),  # type: ignore[arg-type]
            security_groups=self._get_param("SecurityGroups", []),
            user_data=user_data,
            instance_type=params.get("InstanceType"),  # type: ignore[arg-type]
            instance_monitoring=self._get_param("InstanceMonitoring.Enabled", False),
            instance_profile_name=params.get("IamInstanceProfile"),
            spot_price=params.get("SpotPrice"),
            ebs_optimized=self._get_bool_param("EbsOptimized", False),
            associate_public_ip_address=self._get_bool_param(
                "AssociatePublicIpAddress", False
            ),
            block_device_mappings=params.get("BlockDeviceMappings"),  # type: ignore[arg-type]
            instance_id=params.get("InstanceId"),
            metadata_options=params.get("MetadataOptions"),
            classic_link_vpc_id=params.get("ClassicLinkVPCId"),
            classic_link_vpc_security_groups=params.get("ClassicLinkVPCSecurityGroups"),
        )
        return EmptyResult()

    def describe_launch_configurations(self) -> ActionResult:
        names = self._get_param("LaunchConfigurationNames", [])
        all_launch_configurations = (
            self.autoscaling_backend.describe_launch_configurations(names)
        )
        marker = self._get_param("NextToken")
        all_names = [lc.name for lc in all_launch_configurations]
        if marker:
            start = all_names.index(marker) + 1
        else:
            start = 0
        # the default is 100, but using 50 to make testing easier
        max_records = self._get_int_param("MaxRecords") or 50
        launch_configurations_resp = all_launch_configurations[
            start : start + max_records
        ]
        next_token = None
        if len(all_launch_configurations) > start + max_records:
            next_token = launch_configurations_resp[-1].name

        result = {
            "LaunchConfigurations": launch_configurations_resp,
            "NextToken": next_token,
        }
        return ActionResult(result)

    def delete_launch_configuration(self) -> ActionResult:
        launch_configurations_name = self._get_param("LaunchConfigurationName")
        self.autoscaling_backend.delete_launch_configuration(launch_configurations_name)
        return EmptyResult()

    def create_auto_scaling_group(self) -> ActionResult:
        params = self._get_params()
        self.autoscaling_backend.create_auto_scaling_group(
            name=self._get_param("AutoScalingGroupName"),
            availability_zones=self._get_param("AvailabilityZones", []),
            desired_capacity=self._get_int_param("DesiredCapacity"),
            max_size=self._get_int_param("MaxSize"),
            min_size=self._get_int_param("MinSize"),
            instance_id=self._get_param("InstanceId"),
            launch_config_name=self._get_param("LaunchConfigurationName"),
            launch_template=self._get_param("LaunchTemplate", {}),
            mixed_instances_policy=params.get("MixedInstancesPolicy"),
            vpc_zone_identifier=self._get_param("VPCZoneIdentifier"),
            default_cooldown=self._get_int_param("DefaultCooldown"),
            health_check_period=self._get_int_param("HealthCheckGracePeriod"),
            health_check_type=self._get_param("HealthCheckType"),
            load_balancers=self._get_param("LoadBalancerNames", []),
            target_group_arns=self._get_param("TargetGroupARNs", []),
            placement_group=self._get_param("PlacementGroup"),
            termination_policies=self._get_param("TerminationPolicies", []),
            tags=params.get("Tags", []),
            capacity_rebalance=self._get_bool_param("CapacityRebalance", False),
            new_instances_protected_from_scale_in=self._get_bool_param(
                "NewInstancesProtectedFromScaleIn", False
            ),
        )
        return EmptyResult()

    def put_scheduled_update_group_action(self) -> ActionResult:
        self.autoscaling_backend.put_scheduled_update_group_action(
            name=self._get_param("AutoScalingGroupName"),
            desired_capacity=self._get_int_param("DesiredCapacity"),
            max_size=self._get_int_param("MaxSize"),
            min_size=self._get_int_param("MinSize"),
            scheduled_action_name=self._get_param("ScheduledActionName"),
            start_time=self._get_param("StartTime"),
            end_time=self._get_param("EndTime"),
            recurrence=self._get_param("Recurrence"),
            timezone=self._get_param("TimeZone"),
        )
        return EmptyResult()

    def batch_put_scheduled_update_group_action(self) -> ActionResult:
        failed_actions = (
            self.autoscaling_backend.batch_put_scheduled_update_group_action(
                name=self._get_param("AutoScalingGroupName"),
                actions=self._get_param("ScheduledUpdateGroupActions", []),
            )
        )
        result = {"FailedScheduledUpdateGroupActions": failed_actions}
        return ActionResult(result)

    def describe_scheduled_actions(self) -> ActionResult:
        scheduled_actions = self.autoscaling_backend.describe_scheduled_actions(
            autoscaling_group_name=self._get_param("AutoScalingGroupName"),
            scheduled_action_names=self._get_param("ScheduledActionNames", []),
        )
        result = {"ScheduledUpdateGroupActions": scheduled_actions}
        return ActionResult(result)

    def delete_scheduled_action(self) -> ActionResult:
        auto_scaling_group_name = self._get_param("AutoScalingGroupName")
        scheduled_action_name = self._get_param("ScheduledActionName")
        self.autoscaling_backend.delete_scheduled_action(
            auto_scaling_group_name=auto_scaling_group_name,
            scheduled_action_name=scheduled_action_name,
        )
        return EmptyResult()

    def batch_delete_scheduled_action(self) -> ActionResult:
        auto_scaling_group_name = self._get_param("AutoScalingGroupName")
        scheduled_action_names = self._get_param("ScheduledActionNames", [])
        failed_actions = self.autoscaling_backend.batch_delete_scheduled_action(
            auto_scaling_group_name=auto_scaling_group_name,
            scheduled_action_names=scheduled_action_names,
        )
        result = {"FailedScheduledActions": failed_actions}
        return ActionResult(result)

    def describe_scaling_activities(self) -> ActionResult:
        result = {"Activities": []}  # type: ignore[var-annotated]
        return ActionResult(result)

    def attach_instances(self) -> ActionResult:
        group_name = self._get_param("AutoScalingGroupName")
        instance_ids = self._get_param("InstanceIds", [])
        self.autoscaling_backend.attach_instances(group_name, instance_ids)
        return EmptyResult()

    def set_instance_health(self) -> ActionResult:
        instance_id = self._get_param("InstanceId")
        health_status = self._get_param("HealthStatus")
        if health_status not in ["Healthy", "Unhealthy"]:
            raise ValueError("Valid instance health states are: [Healthy, Unhealthy]")
        self.autoscaling_backend.set_instance_health(instance_id, health_status)
        return EmptyResult()

    def detach_instances(self) -> ActionResult:
        group_name = self._get_param("AutoScalingGroupName")
        instance_ids = self._get_param("InstanceIds", [])
        should_decrement = self._get_bool_param("ShouldDecrementDesiredCapacity", False)
        activities = self.autoscaling_backend.detach_instances(
            group_name, instance_ids, should_decrement
        )
        result = {"Activities": activities}
        return ActionResult(result)

    def attach_load_balancer_target_groups(self) -> ActionResult:
        group_name = self._get_param("AutoScalingGroupName")
        target_group_arns = self._get_param("TargetGroupARNs", [])

        self.autoscaling_backend.attach_load_balancer_target_groups(
            group_name, target_group_arns
        )
        return EmptyResult()

    def describe_load_balancer_target_groups(self) -> ActionResult:
        group_name = self._get_param("AutoScalingGroupName")
        target_group_arns = (
            self.autoscaling_backend.describe_load_balancer_target_groups(group_name)
        )
        result = {
            "LoadBalancerTargetGroups": [
                {"LoadBalancerTargetGroupARN": arn, "State": "Added"}
                for arn in target_group_arns
            ]
        }
        return ActionResult(result)

    def detach_load_balancer_target_groups(self) -> ActionResult:
        group_name = self._get_param("AutoScalingGroupName")
        target_group_arns = self._get_param("TargetGroupARNs", [])

        self.autoscaling_backend.detach_load_balancer_target_groups(
            group_name, target_group_arns
        )
        return EmptyResult()

    def describe_auto_scaling_groups(self) -> ActionResult:
        names = self._get_param("AutoScalingGroupNames", [])
        token = self._get_param("NextToken")
        filters = self._get_param("Filters", [])
        all_groups = self.autoscaling_backend.describe_auto_scaling_groups(
            names, filters=filters
        )
        all_names = [group.name for group in all_groups]
        if token:
            start = all_names.index(token) + 1
        else:
            start = 0
        max_records = self._get_int_param("MaxRecords", 50)
        if max_records > 100:
            raise ValueError
        groups = all_groups[start : start + max_records]
        next_token = None
        if max_records and len(all_groups) > start + max_records:
            next_token = groups[-1].name
        result = {"AutoScalingGroups": groups, "NextToken": next_token}
        return ActionResult(result)

    def update_auto_scaling_group(self) -> ActionResult:
        self.autoscaling_backend.update_auto_scaling_group(
            name=self._get_param("AutoScalingGroupName"),
            availability_zones=self._get_param("AvailabilityZones", []),
            desired_capacity=self._get_int_param("DesiredCapacity"),
            max_size=self._get_int_param("MaxSize"),
            min_size=self._get_int_param("MinSize"),
            launch_config_name=self._get_param("LaunchConfigurationName"),
            launch_template=self._get_param("LaunchTemplate", {}),
            vpc_zone_identifier=self._get_param("VPCZoneIdentifier"),
            health_check_period=self._get_int_param("HealthCheckGracePeriod"),
            health_check_type=self._get_param("HealthCheckType"),
            new_instances_protected_from_scale_in=self._get_bool_param(
                "NewInstancesProtectedFromScaleIn", None
            ),
            mixed_instances_policy=self._get_param("MixedInstancesPolicy"),
        )
        return EmptyResult()

    def delete_auto_scaling_group(self) -> ActionResult:
        group_name = self._get_param("AutoScalingGroupName")
        self.autoscaling_backend.delete_auto_scaling_group(group_name)
        return EmptyResult()

    def set_desired_capacity(self) -> ActionResult:
        group_name = self._get_param("AutoScalingGroupName")
        desired_capacity = self._get_int_param("DesiredCapacity")
        self.autoscaling_backend.set_desired_capacity(group_name, desired_capacity)
        return EmptyResult()

    def create_or_update_tags(self) -> ActionResult:
        self.autoscaling_backend.create_or_update_tags(self._get_param("Tags", []))
        return EmptyResult()

    def delete_tags(self) -> ActionResult:
        self.autoscaling_backend.delete_tags(self._get_params().get("Tags", []))
        return EmptyResult()

    def describe_auto_scaling_instances(self) -> ActionResult:
        instance_states = self.autoscaling_backend.describe_auto_scaling_instances(
            instance_ids=self._get_param("InstanceIds", [])
        )
        result = {"AutoScalingInstances": instance_states}
        return ActionResult(result)

    def put_lifecycle_hook(self) -> ActionResult:
        self.autoscaling_backend.create_lifecycle_hook(
            name=self._get_param("LifecycleHookName"),
            as_name=self._get_param("AutoScalingGroupName"),
            transition=self._get_param("LifecycleTransition"),
            timeout=self._get_int_param("HeartbeatTimeout"),
            result=self._get_param("DefaultResult"),
        )
        return EmptyResult()

    def describe_lifecycle_hooks(self) -> ActionResult:
        lifecycle_hooks = self.autoscaling_backend.describe_lifecycle_hooks(
            as_name=self._get_param("AutoScalingGroupName"),
            lifecycle_hook_names=self._get_param("LifecycleHookNames", []),
        )
        result = {"LifecycleHooks": lifecycle_hooks}
        return ActionResult(result)

    def delete_lifecycle_hook(self) -> ActionResult:
        as_name = self._get_param("AutoScalingGroupName")
        name = self._get_param("LifecycleHookName")
        self.autoscaling_backend.delete_lifecycle_hook(as_name, name)
        return EmptyResult()

    def put_scaling_policy(self) -> ActionResult:
        params = self._get_params()
        policy = self.autoscaling_backend.put_scaling_policy(
            name=params.get("PolicyName"),  # type: ignore[arg-type]
            policy_type=params.get("PolicyType", "SimpleScaling"),
            metric_aggregation_type=params.get("MetricAggregationType"),  # type: ignore[arg-type]
            adjustment_type=params.get("AdjustmentType"),  # type: ignore[arg-type]
            as_name=params.get("AutoScalingGroupName"),  # type: ignore[arg-type]
            min_adjustment_magnitude=params.get("MinAdjustmentMagnitude"),  # type: ignore[arg-type]
            scaling_adjustment=self._get_int_param("ScalingAdjustment"),
            cooldown=self._get_int_param("Cooldown"),
            target_tracking_config=params.get("TargetTrackingConfiguration", {}),
            step_adjustments=params.get("StepAdjustments", []),
            estimated_instance_warmup=params.get("EstimatedInstanceWarmup"),  # type: ignore[arg-type]
            predictive_scaling_configuration=params.get(
                "PredictiveScalingConfiguration", {}
            ),
        )
        return ActionResult({"PolicyArn": policy.arn})

    def describe_policies(self) -> ActionResult:
        policies = self.autoscaling_backend.describe_policies(
            autoscaling_group_name=self._get_param("AutoScalingGroupName"),
            policy_names=self._get_param("PolicyNames", []),
            policy_types=self._get_param("PolicyTypes", []),
        )
        result = {"ScalingPolicies": policies}
        return ActionResult(result)

    def delete_policy(self) -> ActionResult:
        group_name = self._get_param("PolicyName")
        self.autoscaling_backend.delete_policy(group_name)
        return EmptyResult()

    def execute_policy(self) -> ActionResult:
        group_name = self._get_param("PolicyName")
        self.autoscaling_backend.execute_policy(group_name)
        return EmptyResult()

    def attach_load_balancers(self) -> ActionResult:
        group_name = self._get_param("AutoScalingGroupName")
        load_balancer_names = self._get_param("LoadBalancerNames", [])
        self.autoscaling_backend.attach_load_balancers(group_name, load_balancer_names)
        return EmptyResult()

    def describe_load_balancers(self) -> ActionResult:
        group_name = self._get_param("AutoScalingGroupName")
        load_balancers = self.autoscaling_backend.describe_load_balancers(group_name)
        result = {
            "LoadBalancers": [
                {"LoadBalancerName": name, "State": "Added"} for name in load_balancers
            ]
        }
        return ActionResult(result)

    def detach_load_balancers(self) -> ActionResult:
        group_name = self._get_param("AutoScalingGroupName")
        load_balancer_names = self._get_param("LoadBalancerNames", [])
        self.autoscaling_backend.detach_load_balancers(group_name, load_balancer_names)
        return EmptyResult()

    def enter_standby(self) -> ActionResult:
        group_name = self._get_param("AutoScalingGroupName")
        instance_ids = self._get_param("InstanceIds", [])
        should_decrement = self._get_bool_param("ShouldDecrementDesiredCapacity")
        activities = self.autoscaling_backend.enter_standby_instances(
            group_name, instance_ids, should_decrement
        )
        result = {"Activities": activities}
        return ActionResult(result)

    def exit_standby(self) -> ActionResult:
        group_name = self._get_param("AutoScalingGroupName")
        instance_ids = self._get_param("InstanceIds", [])
        activities = self.autoscaling_backend.exit_standby_instances(
            group_name, instance_ids
        )
        result = {"Activities": activities}
        return ActionResult(result)

    def suspend_processes(self) -> ActionResult:
        autoscaling_group_name = self._get_param("AutoScalingGroupName")
        scaling_processes = self._get_param("ScalingProcesses", [])
        self.autoscaling_backend.suspend_processes(
            autoscaling_group_name, scaling_processes
        )
        return EmptyResult()

    def resume_processes(self) -> ActionResult:
        autoscaling_group_name = self._get_param("AutoScalingGroupName")
        scaling_processes = self._get_param("ScalingProcesses", [])
        self.autoscaling_backend.resume_processes(
            autoscaling_group_name, scaling_processes
        )
        return EmptyResult()

    def set_instance_protection(self) -> ActionResult:
        group_name = self._get_param("AutoScalingGroupName")
        instance_ids = self._get_param("InstanceIds", [])
        protected_from_scale_in = self._get_bool_param("ProtectedFromScaleIn")
        self.autoscaling_backend.set_instance_protection(
            group_name, instance_ids, protected_from_scale_in
        )
        return EmptyResult()

    def terminate_instance_in_auto_scaling_group(self) -> ActionResult:
        instance_id = self._get_param("InstanceId")
        should_decrement = self._get_bool_param("ShouldDecrementDesiredCapacity", False)
        activity = self.autoscaling_backend.terminate_instance(
            instance_id, should_decrement
        )
        result = {"Activity": activity}
        return ActionResult(result)

    def describe_tags(self) -> ActionResult:
        filters = self._get_param("Filters", [])
        tags = self.autoscaling_backend.describe_tags(filters=filters)
        result = {"Tags": tags, "NextToken": None}
        return ActionResult(result)

    def enable_metrics_collection(self) -> ActionResult:
        group_name = self._get_param("AutoScalingGroupName")
        metrics = self._get_param("Metrics")
        self.autoscaling_backend.enable_metrics_collection(group_name, metrics)  # type: ignore[arg-type]
        return EmptyResult()

    def put_warm_pool(self) -> ActionResult:
        params = self._get_params()
        group_name = params.get("AutoScalingGroupName")
        max_group_prepared_capacity = params.get("MaxGroupPreparedCapacity")
        min_size = params.get("MinSize")
        pool_state = params.get("PoolState")
        instance_reuse_policy = params.get("InstanceReusePolicy")
        self.autoscaling_backend.put_warm_pool(
            group_name=group_name,  # type: ignore[arg-type]
            max_group_prepared_capacity=max_group_prepared_capacity,
            min_size=min_size,
            pool_state=pool_state,
            instance_reuse_policy=instance_reuse_policy,
        )
        return EmptyResult()

    def describe_warm_pool(self) -> ActionResult:
        group_name = self._get_param("AutoScalingGroupName")
        warm_pool = self.autoscaling_backend.describe_warm_pool(group_name=group_name)
        result = {"WarmPoolConfiguration": warm_pool, "Instances": []}  # type: ignore[var-annotated]
        return ActionResult(result)

    def delete_warm_pool(self) -> ActionResult:
        group_name = self._get_param("AutoScalingGroupName")
        self.autoscaling_backend.delete_warm_pool(group_name=group_name)
        return EmptyResult()



from collections.abc import Iterator
from typing import Any, Optional

from moto.acm.models import AWSCertificateManagerBackend, acm_backends
from moto.appsync.models import AppSyncBackend, appsync_backends
from moto.athena.models import athena_backends
from moto.awslambda.models import LambdaBackend, lambda_backends
from moto.backup.models import BackupBackend, backup_backends
from moto.clouddirectory import CloudDirectoryBackend, clouddirectory_backends
from moto.cloudfront.models import CloudFrontBackend, cloudfront_backends
from moto.cloudwatch.models import CloudWatchBackend, cloudwatch_backends
from moto.comprehend.models import ComprehendBackend, comprehend_backends
from moto.connectcampaigns.models import (
    ConnectCampaignServiceBackend,
    connectcampaigns_backends,
)
from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.exceptions import RESTError
from moto.directconnect.models import DirectConnectBackend, directconnect_backends
from moto.dms.models import DatabaseMigrationServiceBackend, dms_backends
from moto.dynamodb.models import DynamoDBBackend, dynamodb_backends
from moto.ec2 import ec2_backends
from moto.ecs.models import EC2ContainerServiceBackend, ecs_backends
from moto.efs.models import EFSBackend, efs_backends
from moto.elasticache.models import ElastiCacheBackend, elasticache_backends
from moto.elasticbeanstalk.models import EBBackend, eb_backends
from moto.elb.models import ELBBackend, elb_backends
from moto.elbv2.models import ELBv2Backend, elbv2_backends
from moto.emr.models import ElasticMapReduceBackend, emr_backends
from moto.events.models import EventsBackend, events_backends
from moto.firehose.models import FirehoseBackend, firehose_backends
from moto.fsx.models import FSxBackend, fsx_backends
from moto.glacier.models import GlacierBackend, glacier_backends
from moto.glue.models import GlueBackend, glue_backends
from moto.kafka.models import KafkaBackend, kafka_backends
from moto.kinesis.models import KinesisBackend, kinesis_backends
from moto.kinesisanalyticsv2.models import (
    KinesisAnalyticsV2Backend,
    kinesisanalyticsv2_backends,
)
from moto.kms.models import KmsBackend, kms_backends
from moto.lexv2models.models import LexModelsV2Backend, lexv2models_backends
from moto.logs.models import LogsBackend, logs_backends
from moto.moto_api._internal import mock_random
from moto.quicksight.models import QuickSightBackend, quicksight_backends
from moto.rds.models import RDSBackend, rds_backends
from moto.redshift.models import RedshiftBackend, redshift_backends
from moto.s3.models import S3Backend, s3_backends
from moto.sagemaker.models import SageMakerModelBackend, sagemaker_backends
from moto.secretsmanager import secretsmanager_backends
from moto.secretsmanager.models import ReplicaSecret, SecretsManagerBackend
from moto.servicecatalog.models import ServiceCatalogBackend, servicecatalog_backends
from moto.sesv2.models import SESV2Backend, sesv2_backends
from moto.sns.models import SNSBackend, sns_backends
from moto.sqs.models import SQSBackend, sqs_backends
from moto.ssm.models import SimpleSystemManagerBackend, ssm_backends
from moto.ssm.utils import parameter_arn
from moto.stepfunctions.models import StepFunctionBackend, stepfunctions_backends
from moto.utilities.tagging_service import TaggingService
from moto.utilities.utils import get_partition
from moto.vpclattice.models import VPCLatticeBackend, vpclattice_backends
from moto.workspaces.models import WorkSpacesBackend, workspaces_backends
from moto.workspacesweb.models import WorkSpacesWebBackend, workspacesweb_backends

# Left: EC2 RDS ELB Lambda EMR Glacier Kinesis Redshift Route53
# StorageGateway DynamoDB MachineLearning ACM DirectConnect DirectoryService CloudHSM
# Inspector Elasticsearch


class ResourceGroupsTaggingAPIBackend(BaseBackend):
    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)

        self._pages: dict[str, Any] = {}
        # Like 'someuuid': {'gen': <generator>, 'misc': None}
        # Misc is there for peeking from a generator and it cant
        # fit in the current request. As we only store generators
        # there is really no point cleaning up

    @property
    def appsync_backend(self) -> AppSyncBackend:
        return appsync_backends[self.account_id][self.region_name]

    @property
    def s3_backend(self) -> S3Backend:
        return s3_backends[self.account_id][self.partition]

    @property
    def directconnect_backend(self) -> DirectConnectBackend:
        return directconnect_backends[self.account_id][self.region_name]

    @property
    def dms_backend(self) -> DatabaseMigrationServiceBackend:
        return dms_backends[self.account_id][self.region_name]

    @property
    def ec2_backend(self) -> Any:  # type: ignore[misc]
        return ec2_backends[self.account_id][self.region_name]

    @property
    def efs_backend(self) -> EFSBackend:
        return efs_backends[self.account_id][self.region_name]

    @property
    def eb_backend(self) -> EBBackend:
        return eb_backends[self.account_id][self.region_name]

    @property
    def elb_backend(self) -> ELBBackend:
        return elb_backends[self.account_id][self.region_name]

    @property
    def elbv2_backend(self) -> ELBv2Backend:
        return elbv2_backends[self.account_id][self.region_name]

    @property
    def events_backend(self) -> EventsBackend:
        return events_backends[self.account_id][self.region_name]

    @property
    def glue_backend(self) -> GlueBackend:
        return glue_backends[self.account_id][self.region_name]

    @property
    def kinesis_backend(self) -> KinesisBackend:
        return kinesis_backends[self.account_id][self.region_name]

    @property
    def kinesisanalyticsv2_backend(self) -> KinesisAnalyticsV2Backend:
        return kinesisanalyticsv2_backends[self.account_id][self.region_name]

    @property
    def kms_backend(self) -> KmsBackend:
        return kms_backends[self.account_id][self.region_name]

    @property
    def logs_backend(self) -> LogsBackend:
        return logs_backends[self.account_id][self.region_name]

    @property
    def rds_backend(self) -> RDSBackend:
        return rds_backends[self.account_id][self.region_name]

    @property
    def fsx_backends(self) -> FSxBackend:
        return fsx_backends[self.account_id][self.region_name]

    @property
    def glacier_backend(self) -> GlacierBackend:
        return glacier_backends[self.account_id][self.region_name]

    @property
    def emr_backend(self) -> ElasticMapReduceBackend:
        return emr_backends[self.account_id][self.region_name]

    @property
    def redshift_backend(self) -> RedshiftBackend:
        return redshift_backends[self.account_id][self.region_name]

    @property
    def lambda_backend(self) -> LambdaBackend:
        return lambda_backends[self.account_id][self.region_name]

    @property
    def ecs_backend(self) -> EC2ContainerServiceBackend:
        return ecs_backends[self.account_id][self.region_name]

    @property
    def firehose_backend(self) -> FirehoseBackend:
        return firehose_backends[self.account_id][self.region_name]

    @property
    def acm_backend(self) -> AWSCertificateManagerBackend:
        return acm_backends[self.account_id][self.region_name]

    @property
    def secretsmanager_backend(self) -> SecretsManagerBackend:
        return secretsmanager_backends[self.account_id][self.region_name]

    @property
    def sns_backend(self) -> SNSBackend:
        return sns_backends[self.account_id][self.region_name]

    @property
    def ssm_backend(self) -> SimpleSystemManagerBackend:
        return ssm_backends[self.account_id][self.region_name]

    @property
    def sqs_backend(self) -> SQSBackend:
        return sqs_backends[self.account_id][self.region_name]

    @property
    def servicecatalog_backend(self) -> ServiceCatalogBackend:
        return servicecatalog_backends[self.account_id][self.region_name]

    @property
    def stepfunctions_backend(self) -> StepFunctionBackend:
        return stepfunctions_backends[self.account_id][self.region_name]

    @property
    def backup_backend(self) -> BackupBackend:
        return backup_backends[self.account_id][self.region_name]

    @property
    def dynamodb_backend(self) -> DynamoDBBackend:
        return dynamodb_backends[self.account_id][self.region_name]

    @property
    def workspaces_backend(self) -> Optional[WorkSpacesBackend]:
        # Workspaces service has limited region availability
        if self.region_name in workspaces_backends[self.account_id].regions:
            return workspaces_backends[self.account_id][self.region_name]
        return None

    @property
    def workspacesweb_backends(self) -> Optional[WorkSpacesWebBackend]:
        # WorkspacesWeb service has limited region availability
        if self.region_name in workspacesweb_backends[self.account_id].regions:
            return workspacesweb_backends[self.account_id][self.region_name]
        return None

    @property
    def comprehend_backend(self) -> Optional[ComprehendBackend]:
        # aws Comprehend has limited region availability
        if self.region_name in comprehend_backends[self.account_id].regions:
            return comprehend_backends[self.account_id][self.region_name]
        return None

    @property
    def kafka_backend(self) -> KafkaBackend:
        return kafka_backends[self.account_id][self.region_name]

    @property
    def sagemaker_backend(self) -> SageMakerModelBackend:
        return sagemaker_backends[self.account_id][self.region_name]

    @property
    def lexv2_backend(self) -> Optional[LexModelsV2Backend]:
        if self.region_name in lexv2models_backends[self.account_id].regions:
            return lexv2models_backends[self.account_id][self.region_name]
        return None

    @property
    def clouddirectory_backend(self) -> Optional[CloudDirectoryBackend]:
        if self.region_name in clouddirectory_backends[self.account_id].regions:
            return clouddirectory_backends[self.account_id][self.region_name]
        return None

    @property
    def cloudfront_backend(self) -> CloudFrontBackend:
        return cloudfront_backends[self.account_id][self.partition]

    @property
    def cloudwatch_backend(self) -> CloudWatchBackend:
        return cloudwatch_backends[self.account_id][self.region_name]

    @property
    def connectcampaigns_backend(self) -> Optional[ConnectCampaignServiceBackend]:
        # Connect Campaigns service has limited region availability
        if self.region_name in connectcampaigns_backends[self.account_id].regions:
            return connectcampaigns_backends[self.account_id][self.region_name]
        return None

    @property
    def quicksight_backend(self) -> Optional[QuickSightBackend]:
        if self.region_name in quicksight_backends[self.account_id].regions:
            return quicksight_backends[self.account_id][self.region_name]
        return None

    @property
    def elasticache_backend(self) -> ElastiCacheBackend:
        return elasticache_backends[self.account_id][self.region_name]

    @property
    def vpclattice_backend(self) -> VPCLatticeBackend:
        return vpclattice_backends[self.account_id][self.region_name]

    @property
    def sesv2_backend(self) -> SESV2Backend:
        return sesv2_backends[self.account_id][self.region_name]

    def _get_resources_generator(
        self,
        tag_filters: Optional[list[dict[str, Any]]] = None,
        resource_type_filters: Optional[list[str]] = None,
    ) -> Iterator[dict[str, Any]]:
        # Look at
        # https://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html

        # TODO move these to their respective backends
        filters = []
        for tag_filter_dict in tag_filters:  # type: ignore
            values = tag_filter_dict.get("Values", [])
            if len(values) == 0:
                # Check key matches
                filters.append(lambda t, v, key=tag_filter_dict["Key"]: t == key)
            elif len(values) == 1:
                # Check it's exactly the same as key, value
                filters.append(
                    lambda t, v, key=tag_filter_dict["Key"], value=values[0]: t == key  # type: ignore
                    and v == value
                )
            else:
                # Check key matches and value is one of the provided values
                filters.append(
                    lambda t, v, key=tag_filter_dict["Key"], vl=values: t == key  # type: ignore
                    and v in vl
                )

        def tag_filter(tag_list: list[dict[str, Any]]) -> bool:
            result = []
            if tag_filters:
                for f in filters:
                    temp_result = []
                    for tag in tag_list:
                        f_result = f(tag["Key"], tag["Value"])  # type: ignore
                        temp_result.append(f_result)
                    result.append(any(temp_result))
                return all(result)
            else:
                return True

        def format_tags(tags: dict[str, Any]) -> list[dict[str, Any]]:
            result = []
            for key, value in tags.items():
                result.append({"Key": key, "Value": value})
            return result

        def format_tag_keys(
            tags: list[dict[str, Any]], keys: list[str]
        ) -> list[dict[str, Any]]:
            result = []
            for tag in tags:
                result.append({"Key": tag[keys[0]], "Value": tag[keys[1]]})
            return result

        # ACM
        if not resource_type_filters or "acm" in resource_type_filters:
            for certificate in self.acm_backend._certificates.values():
                tags = format_tags(certificate.tags)
                if not tags or not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{certificate.arn}", "Tags": tags}

        # AppSync
        if not resource_type_filters or "appsync" in resource_type_filters:
            for api in self.appsync_backend.graphql_apis.values():
                tags = self.appsync_backend.tagger.list_tags_for_resource(api.arn)[
                    "Tags"
                ]
                if not tags or not tag_filter(
                    tags
                ):  # Skip if no tags, or invalid filter
                    continue

                yield {"ResourceARN": f"{api.arn}", "Tags": tags}

        # Athena
        if not resource_type_filters or "athena" in resource_type_filters:
            athena_backend = athena_backends[self.account_id][self.region_name]

            # Capacity Reservations
            for capacity_reservation in athena_backend.capacity_reservations.values():
                tags = athena_backend.tagger.list_tags_for_resource(
                    capacity_reservation.arn
                )["Tags"]
                if not tags or not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{capacity_reservation.arn}", "Tags": tags}

            # Workgroups
            for work_group in athena_backend.work_groups.values():
                tags = athena_backend.tagger.list_tags_for_resource(work_group.arn)[
                    "Tags"
                ]
                if not tags or not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{work_group.arn}", "Tags": tags}

            # Data Catalogs
            for data_catalog in athena_backend.data_catalogs.values():
                tags = athena_backend.tagger.list_tags_for_resource(data_catalog.arn)[
                    "Tags"
                ]
                if not tags or not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{data_catalog.arn}", "Tags": tags}

        # Backup
        if not resource_type_filters or "backup" in resource_type_filters:
            for vault in self.backup_backend.vaults.values():
                tags = self.backup_backend.tagger.list_tags_for_resource(
                    vault.backup_vault_arn
                )["Tags"]
                if not tags or not tag_filter(
                    tags
                ):  # Skip if no tags, or invalid filter
                    continue

                yield {"ResourceARN": f"{vault.backup_vault_arn}", "Tags": tags}

        # Comprehend
        if self.comprehend_backend:
            comprehend_resource_map: dict[str, dict[str, Any]] = {
                "comprehend:document-classification-job": dict(
                    self.comprehend_backend.jobs
                ),
                "comprehend:document-classifier": dict(
                    self.comprehend_backend.classifiers
                ),
                "comprehend:dominant-language-detection-job": dict(
                    self.comprehend_backend.jobs
                ),
                "comprehend:entity-recognizer": dict(
                    self.comprehend_backend.recognizers
                ),
                "comprehend:entities-detection-job": dict(self.comprehend_backend.jobs),
                "comprehend:endpoint": dict(self.comprehend_backend.endpoints),
                "comprehend:events-detection-job": dict(self.comprehend_backend.jobs),
                "comprehend:flywheel": dict(self.comprehend_backend.flywheels),
                "comprehend:key-phrases-detection-job": dict(
                    self.comprehend_backend.jobs
                ),
                "comprehend:pii-entities-detection-job": dict(
                    self.comprehend_backend.jobs
                ),
                "comprehend:sentiment-detection-job": dict(
                    self.comprehend_backend.jobs
                ),
                "comprehend:targeted-sentiment-detection-job": dict(
                    self.comprehend_backend.jobs
                ),
                "comprehend:topic-detection-job": dict(self.comprehend_backend.jobs),
            }
            for resource_type, resource_source in comprehend_resource_map.items():
                if (
                    not resource_type_filters
                    or "comprehend" in resource_type_filters
                    or resource_type in resource_type_filters
                ):
                    for resource in resource_source.values():
                        arn = getattr(resource, "arn", None) or getattr(
                            resource, "job_arn", None
                        )
                        if not arn:
                            continue

                        tags = self.comprehend_backend.tagger.list_tags_for_resource(
                            arn
                        )["Tags"]
                        if not tag_filter(tags):
                            continue
                        yield {"ResourceARN": f"{arn}", "Tags": tags}

        # S3
        if (
            not resource_type_filters
            or "s3" in resource_type_filters
            or "s3:bucket" in resource_type_filters
        ):
            for bucket in self.s3_backend.buckets.values():
                tags = self.s3_backend.tagger.list_tags_for_resource(bucket.arn)["Tags"]
                if not tags or not tag_filter(
                    tags
                ):  # Skip if no tags, or invalid filter
                    continue
                yield {"ResourceARN": bucket.arn, "Tags": tags}

        # Cloud Directory
        if self.clouddirectory_backend:
            if not resource_type_filters or "clouddirectory" in resource_type_filters:
                clouddirectory_backend = clouddirectory_backends[self.account_id][
                    self.region_name
                ]
                for directory in clouddirectory_backend.directories.values():
                    tags = clouddirectory_backend.tagger.list_tags_for_resource(
                        directory.directory_arn
                    )["Tags"]
                    if not tags or not tag_filter(tags):
                        continue
                    yield {"ResourceARN": f"{directory.directory_arn}", "Tags": tags}

        # Firehose
        if self.firehose_backend:
            if not resource_type_filters or "firehose" in resource_type_filters:
                firehose_backend = firehose_backends[self.account_id][self.region_name]
                for delivery_stream in firehose_backend.delivery_streams.values():
                    tags = firehose_backend.tagger.list_tags_for_resource(
                        delivery_stream.delivery_stream_arn
                    )["Tags"]
                    if not tags or not tag_filter(tags):
                        continue
                    yield {
                        "ResourceARN": f"{delivery_stream.delivery_stream_arn}",
                        "Tags": tags,
                    }

        # CloudFormation
        if not resource_type_filters or "cloudformation:stack" in resource_type_filters:
            try:
                from moto.cloudformation import cloudformation_backends

                backend = cloudformation_backends[self.account_id][self.region_name]

                for stack in backend.stacks.values():
                    tags = format_tags(stack.tags)
                    if not tag_filter(tags):
                        continue
                    yield {"ResourceARN": f"{stack.stack_id}", "Tags": tags}

            except Exception:
                pass

        # Cloudfront
        if (
            not resource_type_filters
            or "cloudfront" in resource_type_filters
            or "cloudfront:distributions" in resource_type_filters
        ):
            for dist in self.cloudfront_backend.distributions.values():
                tags = self.cloudfront_backend.tagger.list_tags_for_resource(dist.arn)[
                    "Tags"
                ]
                if (
                    not tag_filter(tags) or len(tags) == 0
                ):  # Skip if no tags, or invalid filter
                    continue
                yield {"ResourceARN": f"{dist.arn}", "Tags": tags}

        if self.cloudwatch_backend:
            cloudwatch_resource_map: dict[str, dict[str, Any]] = {
                "cloudwatch:alarm": self.cloudwatch_backend.alarms,
                "cloudwatch:insight-rule": self.cloudwatch_backend.insight_rules,
            }
            for resource_type, resource_source in cloudwatch_resource_map.items():
                if (
                    not resource_type_filters
                    or "cloudwatch" in resource_type_filters
                    or resource_type in resource_type_filters
                ):
                    for resource in resource_source.values():
                        if resource_type == "cloudwatch:alarm":
                            end_of_arn = f"alarm:{resource.name}"
                        elif resource_type == "cloudwatch:insight-rule":
                            end_of_arn = f"insight-rule/{resource.name}"

                        arn = f"arn:{get_partition(self.region_name)}:cloudwatch:{self.region_name}:{self.account_id}:{end_of_arn}"

                        cw_tags = self.cloudwatch_backend.list_tags_for_resource(arn)

                        tags = format_tags(cw_tags)
                        if not tags or not tag_filter(tags):
                            continue
                        yield {
                            "ResourceARN": arn,
                            "Tags": tags,
                        }

        # Connect Campaigns v1
        if self.connectcampaigns_backend:
            if not resource_type_filters or "connectcampaigns" in resource_type_filters:
                connectcampaigns_backend: ConnectCampaignServiceBackend = (
                    connectcampaigns_backends[self.account_id][self.region_name]
                )
                for campaign in connectcampaigns_backend.campaigns.values():
                    tags = connectcampaigns_backend.tagger.list_tags_for_resource(
                        campaign.arn
                    )["Tags"]
                    if not tags or not tag_filter(tags):
                        continue
                    yield {"ResourceARN": f"{campaign.arn}", "Tags": tags}

        # Direct Connect
        if self.directconnect_backend:
            if not resource_type_filters or "directconnect" in resource_type_filters:
                directconnect_backend = directconnect_backends[self.account_id][
                    self.region_name
                ]

                # Connections
                for connection in directconnect_backend.connections.values():
                    tags = directconnect_backend.tagger.list_tags_for_resource(
                        connection.connection_id
                    )["Tags"]
                    tags = format_tag_keys(tags, ["key", "value"])
                    if not tags or not tag_filter(tags):
                        continue
                    yield {"ResourceARN": f"{connection.connection_id}", "Tags": tags}

                # LAGs
                for lag in directconnect_backend.lags.values():
                    tags = directconnect_backend.tagger.list_tags_for_resource(
                        lag.lag_id
                    )["Tags"]
                    tags = format_tag_keys(tags, ["key", "value"])
                    if not tags or not tag_filter(tags):
                        continue
                    yield {"ResourceARN": f"{lag.lag_id}", "Tags": tags}

        # DMS
        if not resource_type_filters or "dms:endpoint" in resource_type_filters:
            for endpoint in self.dms_backend.endpoints.values():
                tags = self.dms_backend.tagger.list_tags_for_resource(
                    endpoint.endpoint_arn
                )["Tags"]
                if not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{endpoint.endpoint_arn}", "Tags": tags}

        if (
            not resource_type_filters
            or "dms:replication-instance" in resource_type_filters
        ):
            for replication_instance in self.dms_backend.replication_instances.values():
                tags = self.dms_backend.tagger.list_tags_for_resource(
                    replication_instance.arn
                )["Tags"]
                if not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{replication_instance.arn}", "Tags": tags}

        if not resource_type_filters or "dms:task" in resource_type_filters:
            for replication_task in self.dms_backend.replication_tasks.values():
                tags = self.dms_backend.tagger.list_tags_for_resource(
                    replication_task.arn
                )["Tags"]
                if not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{replication_task.arn}", "Tags": tags}

        # ECS
        if not resource_type_filters or "ecs:service" in resource_type_filters:
            for service in self.ecs_backend.services.values():
                tags = format_tag_keys(service.tags, ["key", "value"])
                if not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{service.physical_resource_id}", "Tags": tags}

        if not resource_type_filters or "ecs:cluster" in resource_type_filters:
            for cluster in self.ecs_backend.clusters.values():
                tags = format_tag_keys(cluster.tags, ["key", "value"])  # type: ignore[arg-type]
                if not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{cluster.arn}", "Tags": tags}

        if not resource_type_filters or "ecs:task" in resource_type_filters:
            for task_dict in self.ecs_backend.tasks.values():
                for task in task_dict.values():
                    tags = format_tag_keys(task.tags, ["key", "value"])
                    if not tag_filter(tags):
                        continue
                    yield {"ResourceARN": f"{task.task_arn}", "Tags": tags}

        if not resource_type_filters or "ecs:task-definition" in resource_type_filters:
            for task_definition_dict in self.ecs_backend.task_definitions.values():
                for task_definition in task_definition_dict.values():
                    tags = format_tag_keys(task_definition.tags, ["key", "value"])
                    if not tag_filter(tags):
                        continue
                    yield {"ResourceARN": f"{task_definition.arn}", "Tags": tags}

        # EC2 Resources
        ec2_resource_types = {
            "ec2:image": self.ec2_backend.amis.values(),
            "ec2:instance": (
                instance
                for reservation in self.ec2_backend.reservations.values()
                for instance in reservation.instances
            ),
            "ec2:network-interface": self.ec2_backend.enis.values(),
            "ec2:security-group": (
                sg for vpc in self.ec2_backend.groups.values() for sg in vpc.values()
            ),
            "ec2:snapshot": self.ec2_backend.snapshots.values(),
            "ec2:volume": self.ec2_backend.volumes.values(),
            "ec2:vpc": self.ec2_backend.vpcs.values(),
            "ec2:subnet": (
                subnet
                for subnet in self.ec2_backend.subnets.values()
                for subnet in subnet.values()
            ),
            "ec2:vpc-peering-connection": self.ec2_backend.vpc_pcxs.values(),
            "ec2:transit-gateway": self.ec2_backend.transit_gateways.values(),
            "ec2:transit-gateway-attachment": self.ec2_backend.transit_gateway_attachments.values(),
            "ec2:route-table": self.ec2_backend.route_tables.values(),
            "ec2:customer-gateway": self.ec2_backend.customer_gateways.values(),
            "ec2:vpn-connection": self.ec2_backend.vpn_connections.values(),
            "ec2:natgateway": self.ec2_backend.nat_gateways.values(),
            "ec2:internet-gateway": self.ec2_backend.internet_gateways.values(),
            "ec2:managed-prefix-lists": self.ec2_backend.managed_prefix_lists.values(),
            "ec2:flow-logs": self.ec2_backend.flow_logs.values(),
            "ec2:spot-instance-request": self.ec2_backend.spot_instance_requests.values(),
            # TODO: "ec2:reserved-instance": ...,
        }

        for resource_type, resources in ec2_resource_types.items():
            if (
                not resource_type_filters
                or "ec2" in resource_type_filters
                or resource_type in resource_type_filters
            ):
                for resource in resources:
                    tags = format_tags(self.ec2_backend.tags.get(resource.id, {}))
                    if not tags or not tag_filter(tags):
                        continue
                    resource_type_part = resource_type.split(":", 1)[1]
                    resource_arn = f"arn:{self.partition}:ec2:{self.region_name}:{self.account_id}:{resource_type_part}/{resource.id}"
                    yield {
                        "ResourceARN": resource_arn,
                        "Tags": tags,
                    }

        # EFS, resource type elasticfilesystem:access-point
        if (
            not resource_type_filters
            or "elasticfilesystem" in resource_type_filters
            or "elasticfilesystem:access-point" in resource_type_filters
        ):
            for ap in self.efs_backend.access_points.values():
                tags = self.efs_backend.list_tags_for_resource(ap.access_point_id)
                if not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{ap.access_point_arn}", "Tags": tags}

        # EFS, resource type elasticfilesystem:file-system
        if (
            not resource_type_filters
            or "elasticfilesystem" in resource_type_filters
            or "elasticfilesystem:file-system" in resource_type_filters
        ):
            for fs in self.efs_backend.file_systems_by_id.values():
                tags = self.efs_backend.list_tags_for_resource(fs.file_system_id)
                if not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{fs.file_system_arn}", "Tags": tags}

        elasticache_resource_map: dict[str, dict[str, Any]] = {
            "elasticache:cache_clusters": dict(self.elasticache_backend.cache_clusters),
            "elasticache:replication-group": dict(
                self.elasticache_backend.replication_groups
            ),
            "elasticache:snapshots": dict(self.elasticache_backend.snapshots),
            "elasticache:cache_subnet_groups": dict(
                self.elasticache_backend.cache_subnet_groups
            ),
            "elasticache:users": dict(self.elasticache_backend.users),
        }

        for resource_type, resource_source in elasticache_resource_map.items():
            if (
                not resource_type_filters
                or "elasticache" in resource_type_filters
                or resource_type in resource_type_filters
            ):
                for resource in resource_source.values():
                    if (
                        resource_type == "elasticache:users"
                        and resource.id == "default"
                    ):
                        continue

                    tags = (
                        self.elasticache_backend.tagging_service.list_tags_for_resource(
                            resource.arn
                        )["Tags"]
                    )
                    if not tag_filter(tags):
                        continue
                    yield {"ResourceARN": f"{resource.arn}", "Tags": tags}

        # ElasticBeanstalk
        if (
            not resource_type_filters
            or "elasticbeanstalk:environment" in resource_type_filters
        ):
            for eb_application in self.eb_backend.applications.values():
                for eb_environment in eb_application.environments.values():
                    tags = [
                        {"Key": k, "Value": v} for k, v in eb_environment.tags.items()
                    ]
                    if not tag_filter(tags):
                        continue
                    yield {
                        "ResourceARN": f"{eb_environment.environment_arn}",
                        "Tags": tags,
                    }

        # ELB (Classic Load Balancers)
        if (
            not resource_type_filters
            or "elb" in resource_type_filters
            or "elb:loadbalancer" in resource_type_filters
        ):
            for elb in self.elb_backend.load_balancers.values():
                tags = format_tags(elb.tags)
                if not tags or not tag_filter(
                    tags
                ):  # Skip if no tags, or invalid filter
                    continue

                yield {
                    "ResourceARN": f"arn:{get_partition(self.region_name)}:elasticloadbalancing:{self.region_name}:{self.account_id}:loadbalancer/{elb.name}",
                    "Tags": tags,
                }

        # TODO add these to the keys and values functions / combine functions
        # ELB, resource type elasticloadbalancing:loadbalancer
        if (
            not resource_type_filters
            or "elasticloadbalancing" in resource_type_filters
            or "elasticloadbalancing:loadbalancer" in resource_type_filters
        ):
            for elbv2 in self.elbv2_backend.load_balancers.values():
                tags = self.elbv2_backend.tagging_service.list_tags_for_resource(
                    elbv2.arn
                )["Tags"]
                if not tag_filter(tags):  # Skip if no tags, or invalid filter
                    continue

                yield {"ResourceARN": f"{elbv2.arn}", "Tags": tags}

        # ELB Target Group, resource type elasticloadbalancing:targetgroup
        if (
            not resource_type_filters
            or "elasticloadbalancing" in resource_type_filters
            or "elasticloadbalancing:targetgroup" in resource_type_filters
        ):
            for target_group in self.elbv2_backend.target_groups.values():
                tags = self.elbv2_backend.tagging_service.list_tags_for_resource(
                    target_group.arn
                )["Tags"]
                if not tag_filter(tags):  # Skip if no tags, or invalid filter
                    continue

                yield {"ResourceARN": f"{target_group.arn}", "Tags": tags}

        # EMR Cluster

        # Events
        if (
            not resource_type_filters
            or "events" in resource_type_filters
            or "events:event-bus" in resource_type_filters
        ):
            for bus in self.events_backend.event_buses.values():
                tags = self.events_backend.tagger.list_tags_for_resource(bus.arn)[
                    "Tags"
                ]
                if (
                    not tag_filter(tags) or len(tags) == 0
                ):  # Skip if no tags, or invalid filter
                    continue
                yield {"ResourceARN": f"{bus.arn}", "Tags": tags}

        # FSx
        if not resource_type_filters or "fsx" in resource_type_filters:
            # File system
            for file_system in self.fsx_backends.file_systems.values():
                tags = self.fsx_backends.tagger.list_tags_for_resource(
                    file_system.resource_arn
                )["Tags"]
                if not tags or not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{file_system.resource_arn}", "Tags": tags}

            # Backup
            for backup in self.fsx_backends.backups.values():
                tags = self.fsx_backends.tagger.list_tags_for_resource(
                    backup.resource_arn
                )["Tags"]
                if not tags or not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{backup.resource_arn}", "Tags": tags}

        # Glacier Vault

        # Glue
        if not resource_type_filters or any(
            ("glue" in _type) for _type in resource_type_filters
        ):
            if not resource_type_filters or "glue" in resource_type_filters:
                arns_starting_with = [
                    f"arn:{get_partition(self.region_name)}:glue:{self.region_name}:{self.account_id}:"
                ]
            else:
                arns_starting_with = []
                for resource_type in resource_type_filters:
                    if resource_type.startswith("glue:"):
                        glue_type = resource_type.split(":")[-1]
                        arns_starting_with.append(
                            f"arn:{get_partition(self.region_name)}:glue:{self.region_name}:{self.account_id}:{glue_type}"
                        )
            for glue_arn in self.glue_backend.tagger.tags.keys():
                if any(glue_arn.startswith(arn) for arn in arns_starting_with):
                    tags = self.glue_backend.tagger.list_tags_for_resource(glue_arn)[
                        "Tags"
                    ]
                    yield {"ResourceARN": glue_arn, "Tags": tags}

        # Kinesis

        # KinesisAnalyticsV2
        if self.kinesisanalyticsv2_backend and (
            not resource_type_filters or "kinesisanalyticsv2" in resource_type_filters
        ):
            for application in self.kinesisanalyticsv2_backend.applications.values():
                tags = self.kinesisanalyticsv2_backend.tagger.list_tags_for_resource(
                    application.application_arn
                )["Tags"]
                if not tags or not tag_filter(tags):
                    continue
                yield {
                    "ResourceARN": application.application_arn,
                    "Tags": tags,
                }

        # KMS
        if (
            not resource_type_filters
            or "kms" in resource_type_filters
            or "kms:key" in resource_type_filters
        ):
            for kms_key in self.kms_backend.list_keys():
                tags = format_tag_keys(
                    self.kms_backend.list_resource_tags(kms_key.id).get("Tags", []),
                    ["TagKey", "TagValue"],
                )
                if not tag_filter(tags):  # Skip if no tags, or invalid filter
                    continue

                yield {"ResourceARN": f"{kms_key.arn}", "Tags": tags}

        # LexV2
        if self.lexv2_backend:
            lex_v2_resource_map: dict[str, dict[str, Any]] = {
                "lexv2:bot": self.lexv2_backend.bots,
                "lexv2:bot-alias": self.lexv2_backend.bot_aliases,
            }
            for resource_type, resource_source in lex_v2_resource_map.items():
                if (
                    not resource_type_filters
                    or "lexv2" in resource_type_filters
                    or resource_type in resource_type_filters
                ):
                    for resource in resource_source.values():
                        bot_tags = self.lexv2_backend.list_tags_for_resource(
                            resource.arn
                        )

                        tags = format_tags(bot_tags)
                        if not tags or not tag_filter(tags):
                            continue
                        yield {
                            "ResourceARN": resource.arn,
                            "Tags": tags,
                        }

        # LOGS
        if (
            not resource_type_filters
            or "logs" in resource_type_filters
            or "logs:loggroup" in resource_type_filters
        ):
            for group in self.logs_backend.groups.values():
                log_tags = self.logs_backend.list_tags_for_resource(group.arn)
                tags = format_tags(log_tags)

                if not log_tags or not tag_filter(tags):
                    # Skip if no tags, or invalid filter
                    continue
                yield {"ResourceARN": group.arn, "Tags": tags}

        # Quicksight
        if self.quicksight_backend:
            quicksight_resource_map: dict[str, dict[str, Any]] = {
                "quicksight:dashboards": dict(self.quicksight_backend.dashboards),
                "quicksight:data_sources": dict(self.quicksight_backend.data_sources),
                "quicksight:data_sets": dict(self.quicksight_backend.data_sets),
                "quicksight:users": dict(self.quicksight_backend.users),
            }

            for resource_type, resource_source in quicksight_resource_map.items():
                if (
                    not resource_type_filters
                    or "quicksight" in resource_type_filters
                    or resource_type in resource_type_filters
                ):
                    for resource in resource_source.values():
                        tags = self.quicksight_backend.tagger.list_tags_for_resource(
                            resource.arn
                        )["Tags"]

                        if not tags or not tag_filter(tags):
                            continue

                        yield {
                            "ResourceARN": resource.arn,
                            "Tags": tags,
                        }

        # RDS resources
        resource_map: dict[str, dict[str, Any]] = {
            "rds:cluster": dict(self.rds_backend.clusters),
            "rds:db": dict(self.rds_backend.databases),
            "rds:snapshot": dict(self.rds_backend.database_snapshots),
            "rds:cluster-snapshot": dict(self.rds_backend.cluster_snapshots),
            "rds:db-proxy": self.rds_backend.db_proxies,
        }
        for resource_type, resource_source in resource_map.items():
            if (
                not resource_type_filters
                or "rds" in resource_type_filters
                or resource_type in resource_type_filters
            ):
                for resource in resource_source.values():
                    tags = resource.get_tags()
                    if not tags or not tag_filter(tags):
                        continue
                    yield {
                        "ResourceARN": resource.arn,
                        "Tags": tags,
                    }

        # RDS Reserved Database Instance
        # RDS Option Group
        # RDS Parameter Group
        # RDS Security Group
        # RDS Subnet Group
        # RDS Event Subscription

        # RedShift Cluster
        # RedShift Hardware security module (HSM) client certificate
        # RedShift HSM connection
        # RedShift Parameter group
        # RedShift Snapshot
        # RedShift Subnet group

        # Secrets Manager
        if (
            not resource_type_filters
            or "secretsmanager" in resource_type_filters
            or "secretsmanager:secret" in resource_type_filters
        ):
            for secret in self.secretsmanager_backend.secrets.values():
                if isinstance(secret, ReplicaSecret):
                    secret_tags = secret.source.tags
                else:
                    secret_tags = secret.tags

                if secret_tags:
                    formated_tags = format_tag_keys(secret_tags, ["Key", "Value"])
                    if not formated_tags or not tag_filter(formated_tags):
                        continue
                    yield {"ResourceARN": f"{secret.arn}", "Tags": formated_tags}

        # SQS
        if (
            not resource_type_filters
            or "sqs" in resource_type_filters
            or "sqs:queue" in resource_type_filters
        ):
            for queue in self.sqs_backend.queues.values():
                tags = format_tags(queue.tags)
                if not tags or not tag_filter(
                    tags
                ):  # Skip if no tags, or invalid filter
                    continue

                yield {"ResourceARN": f"{queue.queue_arn}", "Tags": tags}

        # Service Catalog
        if not resource_type_filters or "servicecatalog" in resource_type_filters:
            # Portfolio
            for portfolio in self.servicecatalog_backend.portfolios.values():
                tags = self.servicecatalog_backend.tagger.list_tags_for_resource(
                    portfolio.arn
                )["Tags"]
                if not tags or not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{portfolio.arn}", "Tags": tags}

            # Product
            for product in self.servicecatalog_backend.products.values():
                tags = self.servicecatalog_backend.tagger.list_tags_for_resource(
                    product.arn
                )["Tags"]
                if not tags or not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{product.arn}", "Tags": tags}

        if self.sesv2_backend:
            sesv2_resource_map: dict[str, dict[str, Any]] = {
                "ses:configuration-set": self.sesv2_backend.core_backend.config_sets,
                "ses:contact-list": self.sesv2_backend.core_backend.contacts_lists,
                "ses:dedicated-ip-pool": self.sesv2_backend.core_backend.dedicated_ip_pools,
                "ses:email-identity": self.sesv2_backend.core_backend.email_identities,
            }

            resource_id_attr_map: dict[str, str] = {
                "ses:configuration-set": "configuration_set_name",
                "ses:contact-list": "contact_list_name",
                "ses:dedicated-ip-pool": "pool_name",
                "ses:email-identity": "email_identity",
            }

            for resource_type, resource_source in sesv2_resource_map.items():
                if (
                    not resource_type_filters
                    or "ses" in resource_type_filters
                    or resource_type in resource_type_filters
                ):
                    for resource in resource_source.values():
                        resource_id_attr = resource_id_attr_map[resource_type]
                        resource_id = getattr(resource, resource_id_attr)

                        arn = f"arn:{get_partition(self.region_name)}:ses:{self.region_name}:{self.account_id}:{resource_type.split(':')[-1]}/{resource_id}"

                        tags = self.sesv2_backend.core_backend.tagger.list_tags_for_resource(
                            arn
                        )["Tags"]

                        if not tags or not tag_filter(tags):
                            continue

                        yield {
                            "ResourceARN": arn,
                            "Tags": tags,
                        }
        # SNS
        if not resource_type_filters or "sns" in resource_type_filters:
            for topic in self.sns_backend.topics.values():
                tags = format_tags(topic._tags)
                if not tags or not tag_filter(
                    tags
                ):  # Skip if no tags, or invalid filter
                    continue
                yield {"ResourceARN": f"{topic.arn}", "Tags": tags}

        # SSM Documents
        if (
            not resource_type_filters
            or "ssm" in resource_type_filters
            or "ssm:document" in resource_type_filters
        ):
            for document in self.ssm_backend._documents.values():
                doc_name = document.describe()["Name"]
                tags = self.ssm_backend._get_documents_tags(doc_name)
                if not tags or not tag_filter(
                    tags
                ):  # Skip if no tags, or invalid filter
                    continue
                yield {
                    "ResourceARN": f"arn:{get_partition(self.region_name)}:ssm:{self.region_name}:{self.account_id}:document/{doc_name}",
                    "Tags": tags,
                }

        # SSM Parameters
        if (
            not resource_type_filters
            or "ssm" in resource_type_filters
            or "ssm:parameter" in resource_type_filters
        ):
            for param_name in self.ssm_backend._parameters:
                tags = format_tags(
                    self.ssm_backend.list_tags_for_resource("Parameter", param_name)
                )
                if not tags or not tag_filter(
                    tags
                ):  # Skip if no tags, or invalid filter
                    continue
                yield {
                    "ResourceARN": parameter_arn(
                        self.account_id, self.region_name, param_name
                    ),
                    "Tags": tags,
                }

        # Step Functions
        if not resource_type_filters or "states:stateMachine" in resource_type_filters:
            for state_machine in self.stepfunctions_backend.state_machines:
                tags = format_tag_keys(
                    state_machine.backend.get_tags_list_for_state_machine(
                        state_machine.arn
                    ),
                    [
                        state_machine.backend.tagger.key_name,
                        state_machine.backend.tagger.value_name,
                    ],
                )
                if not tags or not tag_filter(tags):
                    continue
                yield {"ResourceARN": state_machine.arn, "Tags": tags}

        # VPC Lattice
        if not resource_type_filters or "vpc-lattice" in resource_type_filters:
            # Service
            for service in self.vpclattice_backend.services.values():  # type: ignore[assignment]
                tags = self.vpclattice_backend.tagger.list_tags_for_resource(
                    service.arn
                )["Tags"]
                if not tags or not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{service.arn}", "Tags": tags}

            # Service Networks
            for service_network in self.vpclattice_backend.service_networks.values():
                tags = self.vpclattice_backend.tagger.list_tags_for_resource(
                    service_network.arn
                )["Tags"]
                if not tags or not tag_filter(tags):
                    continue
                yield {"ResourceARN": f"{service_network.arn}", "Tags": tags}

        # Workspaces
        if self.workspaces_backend and (
            not resource_type_filters or "workspaces" in resource_type_filters
        ):
            for ws in self.workspaces_backend.workspaces.values():
                tags = format_tag_keys(ws.tags, ["Key", "Value"])
                if not tags or not tag_filter(
                    tags
                ):  # Skip if no tags, or invalid filter
                    continue

                yield {
                    "ResourceARN": f"arn:{get_partition(self.region_name)}:workspaces:{self.region_name}:{self.account_id}:workspace/{ws.workspace_id}",
                    "Tags": tags,
                }

        # Workspace Directories
        if self.workspaces_backend and (
            not resource_type_filters or "workspaces-directory" in resource_type_filters
        ):
            for wd in self.workspaces_backend.workspace_directories.values():
                tags = format_tag_keys(wd.tags, ["Key", "Value"])
                if not tags or not tag_filter(
                    tags
                ):  # Skip if no tags, or invalid filter
                    continue

                yield {
                    "ResourceARN": f"arn:{get_partition(self.region_name)}:workspaces:{self.region_name}:{self.account_id}:directory/{wd.directory_id}",
                    "Tags": tags,
                }

        # Workspace Images
        if self.workspaces_backend and (
            not resource_type_filters or "workspaces-image" in resource_type_filters
        ):
            for wi in self.workspaces_backend.workspace_images.values():
                tags = format_tag_keys(wi.tags, ["Key", "Value"])
                if not tags or not tag_filter(
                    tags
                ):  # Skip if no tags, or invalid filter
                    continue

                yield {
                    "ResourceARN": f"arn:{get_partition(self.region_name)}:workspaces:{self.region_name}:{self.account_id}:workspaceimage/{wi.image_id}",
                    "Tags": tags,
                }

        # Kafka (MSK)
        if self.kafka_backend and (
            not resource_type_filters or "kafka" in resource_type_filters
        ):
            for msk_cluster in self.kafka_backend.clusters.values():
                tag_dict = self.kafka_backend.list_tags_for_resource(msk_cluster.arn)
                tags = [{"Key": key, "Value": value} for key, value in tag_dict.items()]

                if not tags or not tag_filter(tags):
                    continue

                yield {
                    "ResourceARN": msk_cluster.arn,
                    "Tags": tags,
                }

        # Workspaces Web
        if self.workspacesweb_backends and (
            not resource_type_filters or "workspaces-web" in resource_type_filters
        ):
            for portal in self.workspacesweb_backends.portals.values():
                tags = self.workspacesweb_backends.tagger.list_tags_for_resource(
                    portal.arn
                )["Tags"]
                if not tags or not tag_filter(tags):
                    continue
                yield {
                    "ResourceARN": f"arn:{get_partition(self.region_name)}:workspaces-web:{self.region_name}:{self.account_id}:portal/{portal.portal_id}",
                    "Tags": tags,
                }

        # Lambda Instance
        if (
            not resource_type_filters
            or "lambda" in resource_type_filters
            or "lambda:function" in resource_type_filters
        ):
            for f in self.lambda_backend.list_functions():
                tags = format_tags(f.tags)
                if not tags or not tag_filter(tags):
                    continue
                yield {
                    "ResourceARN": f.function_arn,
                    "Tags": tags,
                }

        if (
            not resource_type_filters
            or "dynamodb" in resource_type_filters
            or "dynamodb:table" in resource_type_filters
        ):
            for table in self.dynamodb_backend.tables.values():
                tags = table.tags

                if not tags or not tag_filter(tags):
                    continue
                yield {
                    "ResourceARN": table.table_arn,
                    "Tags": tags,
                }

        # sagemaker cluster, automljob, compilation-job, domain, model-explainability-job-definition, model-quality-job-definition, and hyper-parameter-tuning-job currently supported
        sagemaker_resource_map: dict[str, dict[str, Any]] = {
            "sagemaker:cluster": self.sagemaker_backend.clusters,
            "sagemaker:automl-job": self.sagemaker_backend.auto_ml_jobs,
            "sagemaker:compilation-job": self.sagemaker_backend.compilation_jobs,
            "sagemaker:domain": self.sagemaker_backend.domains,
            "sagemaker:model-explainability-job-definition": self.sagemaker_backend.model_explainability_job_definitions,
            "sagemaker:model-quality-job-definition": self.sagemaker_backend.model_quality_job_definitions,
            "sagemaker:hyper-parameter-tuning-job": self.sagemaker_backend.hyper_parameter_tuning_jobs,
            "sagemaker:model-bias-job-definition": self.sagemaker_backend.model_bias_job_definitions,
            "sagemaker:data-quality-job-definition": self.sagemaker_backend.data_quality_job_definitions,
            "sagemaker:model": self.sagemaker_backend._models,
            "sagemaker:notebook-instance": self.sagemaker_backend.notebook_instances,
            "sagemaker:endpoint-config": self.sagemaker_backend.endpoint_configs,
            "sagemaker:endpoint": self.sagemaker_backend.endpoints,
            "sagemaker:experiment": self.sagemaker_backend.experiments,
            "sagemaker:pipeline": self.sagemaker_backend.pipelines,
            "sagemaker:pipeline-execution": self.sagemaker_backend.pipeline_executions,
            "sagemaker:processing-job": self.sagemaker_backend.processing_jobs,
            "sagemaker:trial": self.sagemaker_backend.trials,
            "sagemaker:trial-component": self.sagemaker_backend.trial_components,
            "sagemaker:training-job": self.sagemaker_backend.training_jobs,
            "sagemaker:transform-job": self.sagemaker_backend.transform_jobs,
            "sagemaker:notebook-instance-lifecycle-config": self.sagemaker_backend.notebook_instance_lifecycle_configurations,
            "sagemaker:model-card": self.sagemaker_backend.model_cards,
            "sagemaker:model-package-group": self.sagemaker_backend.model_package_groups,
            "sagemaker:model-package": self.sagemaker_backend.model_packages,
            "sagemaker:feature-group": self.sagemaker_backend.feature_groups,
        }
        for resource_type, resource_source in sagemaker_resource_map.items():
            if (
                not resource_type_filters
                or "sagemaker" in resource_type_filters
                or resource_type in resource_type_filters
            ):
                for resource in resource_source.values():
                    tags = self.sagemaker_backend.list_tags(resource.arn)[0]
                    if not tags or not tag_filter(tags):
                        continue
                    yield {
                        "ResourceARN": resource.arn,
                        "Tags": tags,
                    }

    def _get_tag_keys_generator(self) -> Iterator[str]:
        # Look at
        # https://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html

        # S3
        for bucket in self.s3_backend.buckets.values():
            tags = self.s3_backend.tagger.get_tag_dict_for_resource(bucket.arn)
            for key, _ in tags.items():
                yield key

        # EC2 tags
        def get_ec2_keys(res_id: str) -> list[dict[str, str]]:
            result = []
            for key in self.ec2_backend.tags.get(res_id, {}):
                result.append(key)
            return result

        # EC2 AMI, resource type ec2:image
        for ami in self.ec2_backend.amis.values():
            for key in get_ec2_keys(ami.id):  # type: ignore[assignment]
                yield key

        # EC2 Instance, resource type ec2:instance
        for reservation in self.ec2_backend.reservations.values():
            for instance in reservation.instances:
                for key in get_ec2_keys(instance.id):  # type: ignore[assignment]
                    yield key

        # EC2 NetworkInterface, resource type ec2:network-interface
        for eni in self.ec2_backend.enis.values():
            for key in get_ec2_keys(eni.id):  # type: ignore[assignment]
                yield key

        # TODO EC2 ReservedInstance

        # EC2 SecurityGroup, resource type ec2:security-group
        for vpc in self.ec2_backend.groups.values():
            for sg in vpc.values():
                for key in get_ec2_keys(sg.id):  # type: ignore[assignment]
                    yield key

        # EC2 Snapshot, resource type ec2:snapshot
        for snapshot in self.ec2_backend.snapshots.values():
            for key in get_ec2_keys(snapshot.id):  # type: ignore[assignment]
                yield key

        # TODO EC2 SpotInstanceRequest

        # EC2 Volume, resource type ec2:volume
        for volume in self.ec2_backend.volumes.values():
            for key in get_ec2_keys(volume.id):  # type: ignore[assignment]
                yield key

        # Glue
        for tag_dict in self.glue_backend.tagger.tags.values():
            yield from tag_dict.keys()

    def _get_tag_values_generator(self, tag_key: str) -> Iterator[str]:
        # Look at
        # https://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html

        # Do S3, resource type s3
        for bucket in self.s3_backend.buckets.values():
            tags = self.s3_backend.tagger.get_tag_dict_for_resource(bucket.arn)
            for key, value in tags.items():
                if key == tag_key:
                    yield value

        # EC2 tags
        def get_ec2_values(res_id: str) -> list[dict[str, str]]:
            result = []
            for key, value in self.ec2_backend.tags.get(res_id, {}).items():
                if key == tag_key:
                    result.append(value)
            return result

        # EC2 AMI, resource type ec2:image
        for ami in self.ec2_backend.amis.values():
            for value in get_ec2_values(ami.id):  # type: ignore[assignment]
                yield value

        # EC2 Instance, resource type ec2:instance
        for reservation in self.ec2_backend.reservations.values():
            for instance in reservation.instances:
                for value in get_ec2_values(instance.id):  # type: ignore[assignment]
                    yield value

        # EC2 NetworkInterface, resource type ec2:network-interface
        for eni in self.ec2_backend.enis.values():
            for value in get_ec2_values(eni.id):  # type: ignore[assignment]
                yield value

        # TODO EC2 ReservedInstance

        # EC2 SecurityGroup, resource type ec2:security-group
        for vpc in self.ec2_backend.groups.values():
            for sg in vpc.values():
                for value in get_ec2_values(sg.id):  # type: ignore[assignment]
                    yield value

        # EC2 Snapshot, resource type ec2:snapshot
        for snapshot in self.ec2_backend.snapshots.values():
            for value in get_ec2_values(snapshot.id):  # type: ignore[assignment]
                yield value

        # TODO EC2 SpotInstanceRequest

        # EC2 Volume, resource type ec2:volume
        for volume in self.ec2_backend.volumes.values():
            for value in get_ec2_values(volume.id):  # type: ignore[assignment]
                yield value

        # Glue
        for tag_dict in self.glue_backend.tagger.tags.values():
            for key, tag_value in tag_dict.items():
                if key == tag_key and tag_value is not None:
                    yield tag_value

    def get_resources(
        self,
        pagination_token: Optional[str] = None,
        resources_per_page: int = 50,
        tags_per_page: int = 100,
        tag_filters: Optional[list[dict[str, Any]]] = None,
        resource_type_filters: Optional[list[str]] = None,
    ) -> tuple[Optional[str], list[dict[str, Any]]]:
        # Simple range checking
        if 100 >= tags_per_page >= 500:
            raise RESTError(
                "InvalidParameterException", "TagsPerPage must be between 100 and 500"
            )
        if 1 >= resources_per_page >= 50:
            raise RESTError(
                "InvalidParameterException", "ResourcesPerPage must be between 1 and 50"
            )

        # If we have a token, go and find the respective generator, or error
        if pagination_token:
            if pagination_token not in self._pages:
                raise RESTError(
                    "PaginationTokenExpiredException", "Token does not exist"
                )

            generator = self._pages[pagination_token]["gen"]
            left_over = self._pages[pagination_token]["misc"]
        else:
            generator = self._get_resources_generator(
                tag_filters=tag_filters, resource_type_filters=resource_type_filters
            )
            left_over = None

        result = []
        current_tags = 0
        current_resources = 0
        if left_over:
            result.append(left_over)
            current_resources += 1
            current_tags += len(left_over["Tags"])

        try:
            while True:
                # Generator format: [{'ResourceARN': str, 'Tags': [{'Key': str, 'Value': str]}, ...]
                next_item = next(generator)
                resource_tags = len(next_item["Tags"])

                if current_resources >= resources_per_page:
                    break
                if current_tags + resource_tags >= tags_per_page:
                    break

                current_resources += 1
                current_tags += resource_tags

                result.append(next_item)

        except StopIteration:
            # Finished generator before invalidating page limiting constraints
            return None, result

        # Didn't hit StopIteration so there's stuff left in generator
        new_token = str(mock_random.uuid4())
        self._pages[new_token] = {"gen": generator, "misc": next_item}

        # Token used up, might as well bin now, if you call it again you're an idiot
        if pagination_token:
            del self._pages[pagination_token]
        return new_token, result

    def get_tag_keys(
        self, pagination_token: Optional[str] = None
    ) -> tuple[Optional[str], list[str]]:
        if pagination_token:
            if pagination_token not in self._pages:
                raise RESTError(
                    "PaginationTokenExpiredException", "Token does not exist"
                )

            generator = self._pages[pagination_token]["gen"]
            left_over = self._pages[pagination_token]["misc"]
        else:
            generator = self._get_tag_keys_generator()
            left_over = None

        result = []
        current_tags = 0
        if left_over:
            result.append(left_over)
            current_tags += 1

        try:
            while True:
                # Generator format: ['tag', 'tag', 'tag', ...]
                next_item = next(generator)

                if current_tags + 1 >= 128:
                    break

                current_tags += 1

                result.append(next_item)

        except StopIteration:
            # Finished generator before invalidating page limiting constraints
            return None, result

        # Didn't hit StopIteration so there's stuff left in generator
        new_token = str(mock_random.uuid4())
        self._pages[new_token] = {"gen": generator, "misc": next_item}

        # Token used up, might as well bin now, if you call it again your an idiot
        if pagination_token:
            del self._pages[pagination_token]

        return new_token, result

    def get_tag_values(
        self, pagination_token: Optional[str], key: str
    ) -> tuple[Optional[str], list[str]]:
        if pagination_token:
            if pagination_token not in self._pages:
                raise RESTError(
                    "PaginationTokenExpiredException", "Token does not exist"
                )

            generator = self._pages[pagination_token]["gen"]
            left_over = self._pages[pagination_token]["misc"]
        else:
            generator = self._get_tag_values_generator(key)
            left_over = None

        result = []
        current_tags = 0
        if left_over:
            result.append(left_over)
            current_tags += 1

        try:
            while True:
                # Generator format: ['value', 'value', 'value', ...]
                next_item = next(generator)

                if current_tags + 1 >= 128:
                    break

                current_tags += 1

                result.append(next_item)

        except StopIteration:
            # Finished generator before invalidating page limiting constraints
            return None, result

        # Didn't hit StopIteration so there's stuff left in generator
        new_token = str(mock_random.uuid4())
        self._pages[new_token] = {"gen": generator, "misc": next_item}

        # Token used up, might as well bin now, if you call it again your an idiot
        if pagination_token:
            del self._pages[pagination_token]

        return new_token, result

    def tag_resources(
        self, resource_arns: list[str], tags: dict[str, str]
    ) -> dict[str, dict[str, Any]]:
        """
        Only CloudFront, DynamoDB, EFS, Elasticache, Lambda Logs, Quicksight RDS, SageMaker, and SES resources are currently supported
        """
        missing_resources = []
        missing_error: dict[str, Any] = {
            "StatusCode": 404,
            "ErrorCode": "InternalServiceException",
            "ErrorMessage": "Service not yet supported",
        }

        for arn in resource_arns:
            if arn.startswith(
                f"arn:{get_partition(self.region_name)}:rds:"
            ) or arn.startswith(f"arn:{get_partition(self.region_name)}:snapshot:"):
                self.rds_backend.add_tags_to_resource(
                    arn, TaggingService.convert_dict_to_tags_input(tags)
                )
            elif arn.startswith(
                f"arn:{get_partition(self.region_name)}:workspaces-web:"
            ):
                resource_id = arn.split("/")[-1]
                self.workspacesweb_backends.create_tags(  # type: ignore[union-attr]
                    resource_id, TaggingService.convert_dict_to_tags_input(tags)
                )
            elif arn.startswith(f"arn:{get_partition(self.region_name)}:workspaces:"):
                resource_id = arn.split("/")[-1]
                self.workspaces_backend.create_tags(  # type: ignore[union-attr]
                    resource_id, TaggingService.convert_dict_to_tags_input(tags)
                )
            elif arn.startswith(f"arn:{get_partition(self.region_name)}:logs:"):
                self.logs_backend.tag_resource(arn, tags)
            elif arn.startswith(f"arn:{get_partition(self.region_name)}:dynamodb"):
                self.dynamodb_backend.tag_resource(
                    arn, TaggingService.convert_dict_to_tags_input(tags)
                )
            elif arn.startswith(f"arn:{get_partition(self.region_name)}:sagemaker:"):
                self.sagemaker_backend.add_tags(
                    arn, TaggingService.convert_dict_to_tags_input(tags)
                )
            elif arn.startswith(f"arn:{get_partition(self.region_name)}:lambda:"):
                self.lambda_backend.tag_resource(arn, tags)
            elif arn.startswith(
                f"arn:{get_partition(self.region_name)}:elasticfilesystem:"
            ):
                resource_id = arn.split("/")[-1]
                self.efs_backend.tag_resource(
                    resource_id, TaggingService.convert_dict_to_tags_input(tags)
                )
            elif arn.startswith(f"arn:{get_partition(self.region_name)}:quicksight:"):
                assert self.quicksight_backend is not None
                self.quicksight_backend.tag_resource(
                    arn, TaggingService.convert_dict_to_tags_input(tags)
                )
            elif arn.startswith(f"arn:{get_partition(self.region_name)}:elasticache:"):
                self.elasticache_backend.add_tags_to_resource(
                    arn, TaggingService.convert_dict_to_tags_input(tags)
                )
            elif arn.startswith(f"arn:{get_partition(self.region_name)}:ses:"):
                self.sesv2_backend.tag_resource(
                    arn, TaggingService.convert_dict_to_tags_input(tags)
                )
            elif arn.startswith(f"arn:{get_partition(self.region_name)}:cloudfront:"):
                self.cloudfront_backend.tag_resource(
                    arn, TaggingService.convert_dict_to_tags_input(tags)
                )
            else:
                missing_resources.append(arn)
        return dict.fromkeys(missing_resources, missing_error)

    def untag_resources(
        self, resource_arn_list: list[str], tag_keys: list[str]
    ) -> dict[str, dict[str, Any]]:
        """
        Only CloudFront, EFS, Elasticache, Lambda, Quicksight, and SES resources are currently supported
        """
        missing_resources = []
        missing_error: dict[str, Any] = {
            "StatusCode": 404,
            "ErrorCode": "InternalServiceException",
            "ErrorMessage": "Service not yet supported",
        }

        for arn in resource_arn_list:
            if arn.startswith(f"arn:{get_partition(self.region_name)}:lambda:"):
                self.lambda_backend.untag_resource(arn, tag_keys)
            elif arn.startswith(
                f"arn:{get_partition(self.region_name)}:elasticfilesystem:"
            ):
                resource_id = arn.split("/")[-1]
                self.efs_backend.untag_resource(resource_id, tag_keys)
            elif arn.startswith(f"arn:{get_partition(self.region_name)}:quicksight:"):
                assert self.quicksight_backend is not None
                self.quicksight_backend.untag_resource(arn, tag_keys)
            elif arn.startswith(f"arn:{get_partition(self.region_name)}:elasticache:"):
                self.elasticache_backend.remove_tags_from_resource(arn, tag_keys)
            elif arn.startswith(f"arn:{get_partition(self.region_name)}:rds:"):
                self.rds_backend.remove_tags_from_resource(arn, tag_keys)
            elif arn.startswith(f"arn:{get_partition(self.region_name)}:ses:"):
                self.sesv2_backend.untag_resource(arn, tag_keys)
            elif arn.startswith(f"arn:{get_partition(self.region_name)}:cloudfront:"):
                self.cloudfront_backend.untag_resource(arn, tag_keys)
            else:
                missing_resources.append(arn)

        return dict.fromkeys(missing_resources, missing_error)


resourcegroupstaggingapi_backends = BackendDict(
    ResourceGroupsTaggingAPIBackend, "resourcegroupstaggingapi"
)
